# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Python implementation of the magnetic mirror descent (MMD) algorithm.

The algorithm operated over the sequence-from with dilated entropy.

See https://arxiv.org/abs/2206.05825.

One iteration of MMD consists of:
1) Compute gradients of dilated entropy
   and payoffs for current sequence form policies.
2) Compute behavioural form policy starting from the bottom
    of the tree and updating gradients of parent nodes along the way.
3) Convert behavioural form policy to equivalent sequence form policy.

The last sequence form policy converges linearly (exponentially fast)
to a \alpha-reduced normal-form  QRE.
"""

import copy
import warnings
import numpy as np
from scipy import stats as scipy_stats

from open_spiel.python import policy
from open_spiel.python.algorithms.sequence_form_utils import _EMPTY_INFOSET_ACTION_KEYS
from open_spiel.python.algorithms.sequence_form_utils import _EMPTY_INFOSET_KEYS
from open_spiel.python.algorithms.sequence_form_utils import _get_action_from_key
from open_spiel.python.algorithms.sequence_form_utils import construct_vars
from open_spiel.python.algorithms.sequence_form_utils import is_root
from open_spiel.python.algorithms.sequence_form_utils import policy_to_sequence
from open_spiel.python.algorithms.sequence_form_utils import sequence_to_policy
from open_spiel.python.algorithms.sequence_form_utils import uniform_random_seq
import pyspiel


def neg_entropy(probs):
  return -scipy_stats.entropy(probs)


def softmax(x):
  unnormalized = np.exp(x - np.max(x))
  return unnormalized / np.sum(unnormalized)


def divergence(x, y, psi_x, psi_y, grad_psi_y):
  """Compute Bregman divergence between x and y, B_psi(x;y).

  Args:
      x: Numpy array.
      y: Numpy array.
      psi_x: Value of psi evaluated at x.
      psi_y: Value of psi evaluated at y.
      grad_psi_y: Gradient of psi evaluated at y.

  Returns:
      Scalar.
  """
  return psi_x - psi_y - np.dot(grad_psi_y, x - y)


def dilated_dgf_divergence(mmd_1, mmd_2):
  """Bregman divergence between two MMDDilatedEnt objects.

      The value is equivalent to a sum of two Bregman divergences
      over the sequence form, one for each player.

  Args:
      mmd_1: MMDDilatedEnt Object
      mmd_2: MMDDilatedEnt Object

  Returns:
      Scalar.
  """

  dgf_values = [mmd_1.dgf_eval(), mmd_2.dgf_eval()]
  dgf_grads = mmd_2.dgf_grads()
  div = 0
  for player in range(2):
    div += divergence(mmd_1.sequences[player], mmd_2.sequences[player],
                      dgf_values[0][player], dgf_values[1][player],
                      dgf_grads[player])
  return div


class MMDDilatedEnt(object):
  r"""Implements Magnetic Mirror Descent (MMD) with Dilated Entropy.

  The implementation uses the sequence form representation.

  The policies converge to a \alpha-reduced normal form QRE of a
  two-player zero-sum extensive-form game. If \alpha is set
  to zero then the method is equivalent to mirror descent ascent
  over the sequence form with dilated entropy and the policies
  will converge on average to a nash equilibrium with
  the appropriate stepsize schedule (or approximate equilirbrium
  for fixed stepsize).

  The main iteration loop is implemented in `update_sequences`:

  ```python
    game = pyspiel.load_game("game_name")
    mmd = MMDDilatedEnt(game, alpha=0.1)
    for i in range(num_iterations):
      mmd.update_sequences()
  ```
  The gap in the regularized game (i.e. 2x exploitability) converges
  to zero and can be computed:

  ```python
      gap = mmd.get_gap()
  ```
  The policy (i.e. behavioural form policy) can be retrieved:
  ```python
      policies = mmd.get_policies()
  ```

  The average sequences and policies can be retrieved:

  ```python
      avg_sequences = mmd.get_avg_sequences()
      avg_policies = mmd.get_avg_policies()
  ```

  """

  empy_state_action_keys = _EMPTY_INFOSET_ACTION_KEYS[:]
  empty_infoset_keys = _EMPTY_INFOSET_KEYS[:]

  def __init__(self, game, alpha, stepsize=None):
    """Initialize the solver object.

    Args:
        game: a zeros-um spiel game with two players.
        alpha: weight of dilated entropy regularization. If alpha > 0 MMD
          will converge to an alpha-QRE. If alpha = 0 mmd will converge to
          Nash on average.
        stepsize: MMD stepsize. Will be set automatically if None.
    """
    assert game.num_players() == 2
    assert game.get_type().utility == pyspiel.GameType.Utility.ZERO_SUM
    assert game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL
    assert (game.get_type().chance_mode
            == pyspiel.GameType.ChanceMode.DETERMINISTIC or
            game.get_type().chance_mode
            == pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC)
    assert alpha >= 0

    self.game = game
    self.alpha = float(alpha)
    (self.infosets, self.infoset_actions_to_seq, self.infoset_action_maps,
     self.infoset_parent_map, self.payoff_mat,
     self.infoset_actions_children) = construct_vars(game)

    if stepsize is not None:
      self.stepsize = stepsize
    else:
      self.stepsize = self.alpha / (np.max(np.abs(self.payoff_mat))**2)

    if self.stepsize == 0.:
      warnings.warn("MMD stepsize is 0, probably because alpha = 0.")

    self.sequences = uniform_random_seq(game, self.infoset_actions_to_seq)
    self.avg_sequences = copy.deepcopy(self.sequences)
    self.iteration_count = 1

  def get_parent_seq(self, player, infostate):
    """Looks up the parent sequence value for a given infostate.

    Args:
        player: player number, either 0 or 1.
        infostate: infostate id string.

    Returns:
        Scalar.
    """
    parent_isa_key = self.infoset_parent_map[player][infostate]
    seq_id = self.infoset_actions_to_seq[player][parent_isa_key]
    parent_seq = self.sequences[player][seq_id]
    return parent_seq

  def get_infostate_seq(self, player, infostate):
    """Gets vector of sequence form values corresponding to a given infostate.

    Args:
        player: player number, either 0 or 1.
        infostate: infostate id string.

    Returns:
        Numpy array.
    """
    seq_idx = [
        self.infoset_actions_to_seq[player][isa_key]
        for isa_key in self.infoset_action_maps[player][infostate]
    ]
    seqs = np.array([self.sequences[player][idx] for idx in seq_idx])
    return seqs

  def dgf_eval(self):
    """Computes the value of dilated entropy for current sequences.

    Returns:
        List of values, one for each player.
    """
    dgf_value = [0., 0.]

    for player in range(2):
      for infostate in self.infosets[player]:

        if is_root(infostate):
          continue

        parent_seq = self.get_parent_seq(player, infostate)
        if parent_seq > 0:
          children_seq = self.get_infostate_seq(player, infostate)
          dgf_value[player] += parent_seq * neg_entropy(
              children_seq / parent_seq)

    return dgf_value

  def dgf_grads(self):
    """Computes gradients of dilated entropy for each player and current seqs.

    Returns:
        A list of numpy arrays.
    """
    grads = [np.zeros(len(self.sequences[0])), np.zeros(len(self.sequences[1]))]
    for player in range(2):
      for infostate in self.infosets[player]:

        # infostates contain empty sequence for root variable
        if is_root(infostate):
          continue

        parent_seq = self.get_parent_seq(player, infostate)
        if parent_seq > 0:

          for isa_key in self.infoset_action_maps[player][infostate]:
            # compute infostate term
            seq_idx = self.infoset_actions_to_seq[player][isa_key]
            seq = self.sequences[player][seq_idx]
            grads[player][seq_idx] += np.log(seq / parent_seq) + 1

            # compute terms from children if there are any
            num_children = len(self.infoset_actions_children[player].get(
                isa_key, []))
            grads[player][seq_idx] -= num_children
    return grads

  def update_sequences(self):
    """Performs one step of MMD."""
    self.iteration_count += 1
    psi_grads = self.dgf_grads()
    # pylint: disable=invalid-unary-operand-type
    grads = [
        (self.stepsize * self.payoff_mat @ self.sequences[1] - psi_grads[0]) /
        ((1 + self.stepsize * self.alpha)),
        (-self.stepsize * self.payoff_mat.T @ self.sequences[0] - psi_grads[1])
        / (1 + self.stepsize * self.alpha)
    ]

    new_policy = policy.TabularPolicy(self.game)
    for player in range(2):
      self._update_state_sequences(self.empty_infoset_keys[player],
                                   grads[player], player, new_policy)

    self.sequences = policy_to_sequence(self.game, new_policy,
                                        self.infoset_actions_to_seq)
    self.update_avg_sequences()

  def _update_state_sequences(self, infostate, g, player, pol):
    """Update the state sequences."""

    isa_keys = self.infoset_action_maps[player][infostate]
    seq_idx = [
        self.infoset_actions_to_seq[player][isa_key] for isa_key in isa_keys
    ]

    for isa_key, isa_idx in zip(isa_keys, seq_idx):

      # update children first if there are any
      children = self.infoset_actions_children[player].get(isa_key, [])
      for child in children:
        self._update_state_sequences(child, g, player, pol)
        # update gradient
        child_isa_keys = self.infoset_action_maps[player][child]
        child_seq_idx = [
            self.infoset_actions_to_seq[player][child_isa_key]
            for child_isa_key in child_isa_keys
        ]
        g_child = np.array([g[idx] for idx in child_seq_idx])

        actions_child = [
            _get_action_from_key(child_isa_key)
            for child_isa_key in child_isa_keys
        ]
        policy_child = pol.policy_for_key(child)[:]
        policy_child = np.array([policy_child[a] for a in actions_child])
        g[isa_idx] += np.dot(g_child, policy_child)
        g[isa_idx] += neg_entropy(policy_child)

    # no update needed for empty sequence
    if is_root(infostate):
      return

    state_policy = pol.policy_for_key(infostate)
    g_infostate = np.array([g[idx] for idx in seq_idx])
    actions = [_get_action_from_key(isa_key) for isa_key in isa_keys]
    new_state_policy = softmax(-g_infostate)
    for action, pr in zip(actions, new_state_policy):
      state_policy[action] = pr

  def get_gap(self):
    """Computes saddle point gap of the regularized game.

    The gap measures convergence to the alpha-QRE.

    Returns:
        Scalar.
    """
    assert self.alpha > 0, "gap cannot be computed for alpha = 0"
    grads = [(self.payoff_mat @ self.sequences[1]) / (self.alpha),
             (-self.payoff_mat.T @ self.sequences[0]) / (self.alpha)]
    dgf_values = self.dgf_eval()

    br_policy = policy.TabularPolicy(self.game)
    for player in range(2):
      self._update_state_sequences(self.empty_infoset_keys[player],
                                   grads[player], player, br_policy)

    br_sequences = policy_to_sequence(self.game, br_policy,
                                      self.infoset_actions_to_seq)
    curr_sequences = copy.deepcopy(self.sequences)
    self.sequences = br_sequences
    br_dgf_values = self.dgf_eval()
    self.sequences = curr_sequences

    # gap of sequences (x,y)
    # d(x) + max_y' x.T A y'-d(y') + d(y) - min_x' d(x') + x'.T Ay

    gap = 0
    gap += curr_sequences[0].T @ self.payoff_mat @ br_sequences[1]
    gap += self.alpha * (dgf_values[1] - br_dgf_values[1])
    gap += self.alpha * (dgf_values[0] - br_dgf_values[0])
    gap += -br_sequences[0].T @ self.payoff_mat @ curr_sequences[1]
    return gap

  def update_avg_sequences(self):
    for player in range(2):
      self.avg_sequences[player] = self.avg_sequences[player] * (
          self.iteration_count - 1) + self.sequences[player]
      self.avg_sequences[
          player] = self.avg_sequences[player] / self.iteration_count

  def current_sequences(self):
    """Retrieves the current sequences.

    Returns:
      the current sequences for each player as list of numpy arrays.
    """
    return self.sequences

  def get_avg_sequences(self):
    """Retrieves the average sequences.

    Returns:
      the average sequences for each player as list of numpy arrays.
    """
    return self.avg_sequences

  def get_policies(self):
    """Convert current sequences to equivalent behavioural form policies.

    Returns:
        spiel TabularPolicy Object.
    """
    return sequence_to_policy(self.sequences, self.game,
                              self.infoset_actions_to_seq,
                              self.infoset_action_maps)

  def get_avg_policies(self):
    """Convert average sequences to equivalent behavioural form policies.

    Returns:
        spiel TabularPolicy Object.
    """
    return sequence_to_policy(self.avg_sequences, self.game,
                              self.infoset_actions_to_seq,
                              self.infoset_action_maps)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Projected Replicator Dynamics Algorithm.

This is an N-player implementation of the Projected Replicator Dynamics
algorithm described in Lanctot et al., 2017: https://arxiv.org/abs/1711.00832.
"""

import numpy as np

from open_spiel.python.algorithms import nfg_utils


def _partial_multi_dot(player_payoff_tensor, strategies, index_avoided):
  """Computes a generalized dot product avoiding one dimension.

  This is used to directly get the expected return of a given action, given
  other players' strategies, for the player indexed by index_avoided.
  Note that the numpy.dot function is used to compute this product, as it ended
  up being (Slightly) faster in performance tests than np.tensordot. Using the
  reduce function proved slower for both np.dot and np.tensordot.

  Args:
    player_payoff_tensor: payoff tensor for player[index_avoided], of dimension
      (dim(vector[0]), dim(vector[1]), ..., dim(vector[-1])).
    strategies: Meta strategy probabilities for each player.
    index_avoided: Player for which we do not compute the dot product.

  Returns:
    Vector of expected returns for each action of player [the player indexed by
      index_avoided].
  """
  new_axis_order = [index_avoided] + [
      i for i in range(len(strategies)) if (i != index_avoided)
  ]
  accumulator = np.transpose(player_payoff_tensor, new_axis_order)
  for i in range(len(strategies) - 1, -1, -1):
    if i != index_avoided:
      accumulator = np.dot(accumulator, strategies[i])
  return accumulator


def _project_distribution(updated_strategy, gamma):
  """Projects the distribution in updated_x to have minimal probabilities.

  Minimal probabilities are set as gamma, and the probabilities are then
  renormalized to sum to 1.

  Args:
    updated_strategy: New distribution value after being updated by update rule.
    gamma: minimal probability value when divided by number of actions.

  Returns:
    Projected distribution.
  """
  # Epsilon approximation of L2-norm projection onto the Delta_gamma space.
  updated_strategy[updated_strategy < gamma] = gamma
  updated_strategy = updated_strategy / np.sum(updated_strategy)
  return updated_strategy


def _approx_simplex_projection(updated_strategy, gamma=0.0):
  """Approximately projects the distribution in updated_x to have minimal probabilities.

  Minimal probabilities are set as gamma, and the probabilities are then
  renormalized to sum to 1.

  Args:
    updated_strategy: New distribution value after being updated by update rule.
    gamma: minimal probability value when divided by number of actions.

  Returns:
    Projected distribution.
  """
  # Epsilon approximation of L2-norm projection onto the Delta_gamma space.
  updated_strategy[updated_strategy < gamma] = gamma
  updated_strategy = updated_strategy / np.sum(updated_strategy)
  return updated_strategy


def _simplex_projection(updated_strategy, gamma=0.0):
  """Project updated_strategy on the closest point in L2-norm on gamma-simplex.

  Based on: https://eng.ucmerced.edu/people/wwang5/papers/SimplexProj.pdf

  Args:
    updated_strategy: New distribution value after being updated by update rule.
    gamma: minimal probability value when divided by number of actions.

  Returns:
    Projected distribution

  Algorithm description:
  It aims to find a scalar lam to be substracted by each dimension of v
  with the restriction that the resulted quantity should lie in [gamma, 1]
  until the resulted vector summed up to 1
  Example: [0.4, 0.7, 0.6], 0.2 -- > find lam=0.25
            --> [max(0.4-0.25, 0.2), max(0.7-0.25, 0.2), max(0.6-0.25, 0.2)]
            --> [0.2,  0.45, 0.35]
  """

  n = len(updated_strategy)
  idx = np.arange(1, n + 1)
  u = np.sort(updated_strategy)[::-1]
  u_tmp = (1 - np.cumsum(u) - (n - idx) * gamma) / idx
  rho = np.searchsorted(u + u_tmp <= gamma, True)
  return np.maximum(updated_strategy + u_tmp[rho - 1], gamma)


def _projected_replicator_dynamics_step(payoff_tensors, strategies, dt, gamma,
                                        use_approx=False):
  """Does one step of the projected replicator dynamics algorithm.

  Args:
    payoff_tensors: List of payoff tensors for each player.
    strategies: List of the strategies used by each player.
    dt: Update amplitude term.
    gamma: Minimum exploratory probability term.
    use_approx: use approximate simplex projection.

  Returns:
    A list of updated strategies for each player.
  """

  # TODO(author4): Investigate whether this update could be fully vectorized.
  new_strategies = []
  for player in range(len(payoff_tensors)):
    current_payoff_tensor = payoff_tensors[player]
    current_strategy = strategies[player]

    values_per_strategy = _partial_multi_dot(current_payoff_tensor, strategies,
                                             player)
    average_return = np.dot(values_per_strategy, current_strategy)
    delta = current_strategy * (values_per_strategy - average_return)

    updated_strategy = current_strategy + dt * delta
    updated_strategy = (
        _approx_simplex_projection(updated_strategy, gamma) if use_approx
        else _simplex_projection(updated_strategy, gamma))
    new_strategies.append(updated_strategy)
  return new_strategies


def projected_replicator_dynamics(payoff_tensors,
                                  prd_initial_strategies=None,
                                  prd_iterations=int(1e5),
                                  prd_dt=1e-3,
                                  prd_gamma=1e-6,
                                  average_over_last_n_strategies=None,
                                  use_approx=False,
                                  **unused_kwargs):
  """The Projected Replicator Dynamics algorithm.

  Args:
    payoff_tensors: List of payoff tensors for each player.
    prd_initial_strategies: Initial list of the strategies used by each player,
      if any. Could be used to speed up the search by providing a good initial
      solution.
    prd_iterations: Number of algorithmic steps to take before returning an
      answer.
    prd_dt: Update amplitude term.
    prd_gamma: Minimum exploratory probability term.
    average_over_last_n_strategies: Running average window size for average
      policy computation. If None, use the whole trajectory.
    use_approx: use the approximate simplex projection.
    **unused_kwargs: Convenient way of exposing an API compatible with other
      methods with possibly different arguments.

  Returns:
    PRD-computed strategies.
  """
  number_players = len(payoff_tensors)
  # Number of actions available to each player.
  action_space_shapes = payoff_tensors[0].shape

  # If no initial starting position is given, start with uniform probabilities.
  new_strategies = prd_initial_strategies or [
      np.ones(action_space_shapes[k]) / action_space_shapes[k]
      for k in range(number_players)
  ]

  averager = nfg_utils.StrategyAverager(number_players, action_space_shapes,
                                        average_over_last_n_strategies)
  averager.append(new_strategies)

  for _ in range(prd_iterations):
    new_strategies = _projected_replicator_dynamics_step(
        payoff_tensors, new_strategies, prd_dt, prd_gamma, use_approx)
    averager.append(new_strategies)
  return averager.average_strategies()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.cfr."""

import itertools

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import cfr_br
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.algorithms import exploitability
import pyspiel

_KUHN_GAME = pyspiel.load_game("kuhn_poker")
_LEDUC_GAME = pyspiel.load_game("leduc_poker")

_KUHN_UNIFORM_POLICY = policy.TabularPolicy(_KUHN_GAME)
_LEDUC_UNIFORM_POLICY = policy.TabularPolicy(_LEDUC_GAME)
_EXPECTED_EXPLOITABILITIES_CFRBR_KUHN = [
    0.9166666666666666, 0.33333333333333337, 0.3194444444444445,
    0.2604166666666667, 0.22666666666666674
]
_EXPECTED_EXPLOITABILITIES_CFRBR_LEDUC = [
    4.747222222222222, 4.006867283950617, 3.4090489231017034,
    2.8982539553095172, 2.5367193593344504
]


class CFRBRTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(
      list(itertools.product([True, False], [True, False])))
  def test_policy_zero_is_uniform(self, linear_averaging, regret_matching_plus):
    game = pyspiel.load_game("leduc_poker")
    cfr_solver = cfr_br.CFRBRSolver(
        game,
        regret_matching_plus=regret_matching_plus,
        linear_averaging=linear_averaging)

    np.testing.assert_array_equal(
        _LEDUC_UNIFORM_POLICY.action_probability_array,
        cfr_solver.current_policy().action_probability_array)
    np.testing.assert_array_equal(
        _LEDUC_UNIFORM_POLICY.action_probability_array,
        cfr_solver.average_policy().action_probability_array)

  def test_policy_and_average_policy(self):
    game = pyspiel.load_game("kuhn_poker")
    cfrbr_solver = cfr_br.CFRBRSolver(game)
    for _ in range(300):
      cfrbr_solver.evaluate_and_update_policy()
    average_policy = cfrbr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2)
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3)

    cfrbr_solver.current_policy()

  @parameterized.parameters([
      (_KUHN_GAME, pyspiel.CFRBRSolver, _EXPECTED_EXPLOITABILITIES_CFRBR_KUHN),
      (_KUHN_GAME, cfr_br.CFRBRSolver, _EXPECTED_EXPLOITABILITIES_CFRBR_KUHN),
      (_LEDUC_GAME, pyspiel.CFRBRSolver,
       _EXPECTED_EXPLOITABILITIES_CFRBR_LEDUC),
      (_LEDUC_GAME, cfr_br.CFRBRSolver, _EXPECTED_EXPLOITABILITIES_CFRBR_LEDUC),
  ])
  def test_cpp_and_python_cfr_br(self, game, solver_cls,
                                 expected_exploitability):
    solver = solver_cls(game)
    for step in range(5):
      solver.evaluate_and_update_policy()

      # We do not compare the policy directly as we do not have an easy way to
      # convert one to the other, so we use the exploitability as a proxy.
      avg_policy = solver.average_policy()
      if solver_cls == pyspiel.CFRBRSolver:
        exploitability_ = pyspiel.nash_conv(game, avg_policy)
      else:
        exploitability_ = exploitability.nash_conv(game, avg_policy)

      self.assertEqual(expected_exploitability[step], exploitability_)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main file to train and evaluate meta-regret and regret matching agents."""

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.examples.meta_cfr.matrix_games import evaluation
from open_spiel.python.examples.meta_cfr.matrix_games import matrix_dataset
from open_spiel.python.examples.meta_cfr.matrix_games import meta_selfplay_agent
from open_spiel.python.examples.meta_cfr.matrix_games import regret_matching_agent


FLAGS = flags.FLAGS
flags.DEFINE_integer("batch_size", 1, "Batch size.")
flags.DEFINE_integer("evaluation_steps", 1000, "Number of evaluation steps.")
flags.DEFINE_integer("num_batches", 1,
                     "Number of batches to train a meta optimizer.")
flags.DEFINE_integer("repeats", 10,
                     "Number of training each batch in meta learning.")
flags.DEFINE_integer("seed", 10, "random seed.")
flags.DEFINE_integer("min_val", 0,
                     "minimum value for randomizing a payoff matrix.")
flags.DEFINE_integer("max_val", 10,
                     "maximum value for randomizing a payoff matrix.")
flags.DEFINE_integer("num_actions", 3, "Number of actions an agent can take.")
flags.DEFINE_bool("single_problem", False,
                  "If the matrix dataset generates only a single matrix.")


def selfplay_main(argv):
  """Self play."""
  del argv
  np.random.seed(FLAGS.seed)
  # rock-paper-scissor
  base_matrix = np.array([[[0, -1, 1], [1, 0, -1], [-1, 1, 0]]] *
                         FLAGS.batch_size)
  dataset = matrix_dataset.Dataset(
      base_matrix=base_matrix,
      num_training_batches=FLAGS.num_batches,
      minval=FLAGS.min_val,
      maxval=FLAGS.max_val)
  data_loader = dataset.get_training_batch()
  eval_payoff_batch = dataset.get_eval_batch()

  mr_agent = meta_selfplay_agent.MetaSelfplayAgent(
      repeats=FLAGS.repeats,
      training_epochs=FLAGS.evaluation_steps,
      data_loader=data_loader)
  mr_agent.train()

  mr_agent2 = meta_selfplay_agent.MetaSelfplayAgent(
      repeats=FLAGS.repeats,
      training_epochs=FLAGS.evaluation_steps,
      data_loader=data_loader)
  mr_agent2.train()

  rm_agent = regret_matching_agent.RegretMatchingAgent(
      num_actions=FLAGS.num_actions, data_loader=data_loader)
  rm_agent.train()

  rm_agent2 = regret_matching_agent.RegretMatchingAgent(
      num_actions=FLAGS.num_actions, data_loader=data_loader)
  rm_agent2.train()

  print("Regret matching")
  evaluation.evaluate_in_selfplay(
      agent_x=rm_agent,
      agent_y=rm_agent2,
      payoff_batch=eval_payoff_batch,
      steps_count=FLAGS.evaluation_steps)

  print("Meta regret matching")
  evaluation.evaluate_in_selfplay(
      agent_x=mr_agent,
      agent_y=mr_agent2,
      payoff_batch=eval_payoff_batch,
      steps_count=FLAGS.evaluation_steps)


if __name__ == "__main__":
  app.run(selfplay_main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils function for routing game experiment."""
# pylint:disable=too-many-lines,import-error,missing-function-docstring,protected-access,too-many-locals,invalid-name,too-many-arguments,too-many-branches,missing-class-docstring,too-few-public-methods

# pylint:disable=line-too-long
import random
import time

import matplotlib.pyplot as plt
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import policy as policy_module
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import external_sampling_mccfr as external_mccfr
from open_spiel.python.algorithms import fictitious_play
from open_spiel.python.algorithms import nfsp
from open_spiel.python.algorithms import noisy_policy
from open_spiel.python.games import dynamic_routing
from open_spiel.python.games import dynamic_routing_utils
from open_spiel.python.mfg.algorithms import distribution as distribution_module
from open_spiel.python.mfg.algorithms import fictitious_play as mean_field_fictitious_play_module
from open_spiel.python.mfg.algorithms import mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv as nash_conv_module
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import dynamic_routing as mean_field_routing_game
import pyspiel
# pylint:enable=line-too-long


def create_games(origin,
                 destination,
                 num_vehicles,
                 graph,
                 max_time_step,
                 time_step_length=1.0,
                 departure_time=None):
  if departure_time is not None:
    raise NotImplementedError("To do.")
  list_of_vehicles = [
      dynamic_routing_utils.Vehicle(origin, destination)
      for _ in range(num_vehicles)
  ]
  game = dynamic_routing.DynamicRoutingGame(
      {
          "max_num_time_step": max_time_step,
          "time_step_length": time_step_length
      },
      network=graph,
      vehicles=list_of_vehicles)
  seq_game = pyspiel.convert_to_turn_based(game)
  od_demand = [
      dynamic_routing_utils.OriginDestinationDemand(origin, destination, 0,
                                                    num_vehicles)
  ]
  mfg_game = mean_field_routing_game.MeanFieldRoutingGame(
      {
          "max_num_time_step": max_time_step,
          "time_step_length": time_step_length
      },
      network=graph,
      od_demand=od_demand)
  return game, seq_game, mfg_game


def create_braess_network(capacity):
  graph_dict = {
      "A": {
          "connection": {
              "B": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0
              }
          },
          "location": [0, 0]
      },
      "B": {
          "connection": {
              "C": {
                  "a": 1.0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 1.0
              },
              "D": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 2.0
              }
          },
          "location": [1, 0]
      },
      "C": {
          "connection": {
              "D": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0.25
              },
              "E": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 2.0
              }
          },
          "location": [2, 1]
      },
      "D": {
          "connection": {
              "E": {
                  "a": 1,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 1.0
              }
          },
          "location": [2, -1]
      },
      "E": {
          "connection": {
              "F": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0.0
              }
          },
          "location": [3, 0]
      },
      "F": {
          "connection": {},
          "location": [4, 0]
      }
  }
  adjacency_list = {
      key: list(value["connection"].keys())
      for key, value in graph_dict.items()
  }
  bpr_a_coefficient = {}
  bpr_b_coefficient = {}
  capacity = {}
  free_flow_travel_time = {}
  for o_node, value_dict in graph_dict.items():
    for d_node, section_dict in value_dict["connection"].items():
      road_section = dynamic_routing_utils._road_section_from_nodes(
          origin=o_node, destination=d_node)
      bpr_a_coefficient[road_section] = section_dict["a"]
      bpr_b_coefficient[road_section] = section_dict["b"]
      capacity[road_section] = section_dict["capacity"]
      free_flow_travel_time[road_section] = section_dict[
          "free_flow_travel_time"]
  node_position = {key: value["location"] for key, value in graph_dict.items()}
  return dynamic_routing_utils.Network(
      adjacency_list,
      node_position=node_position,
      bpr_a_coefficient=bpr_a_coefficient,
      bpr_b_coefficient=bpr_b_coefficient,
      capacity=capacity,
      free_flow_travel_time=free_flow_travel_time)


def create_augmented_braess_network(capacity):
  graph_dict = {
      "A": {
          "connection": {
              "B": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0
              }
          },
          "location": [0, 0]
      },
      "B": {
          "connection": {
              "C": {
                  "a": 1.0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 1.0
              },
              "D": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 2.0
              }
          },
          "location": [1, 0]
      },
      "C": {
          "connection": {
              "D": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0.25
              },
              "E": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 2.0
              }
          },
          "location": [2, 1]
      },
      "D": {
          "connection": {
              "E": {
                  "a": 1,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 1.0
              },
              "G": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0.0
              }
          },
          "location": [2, -1]
      },
      "E": {
          "connection": {
              "F": {
                  "a": 0,
                  "b": 1.0,
                  "capacity": capacity,
                  "free_flow_travel_time": 0.0
              }
          },
          "location": [3, 0]
      },
      "F": {
          "connection": {},
          "location": [4, 0]
      },
      "G": {
          "connection": {},
          "location": [3, -1]
      }
  }
  adjacency_list = {
      key: list(value["connection"].keys())
      for key, value in graph_dict.items()
  }
  bpr_a_coefficient = {}
  bpr_b_coefficient = {}
  capacity = {}
  free_flow_travel_time = {}
  for o_node, value_dict in graph_dict.items():
    for d_node, section_dict in value_dict["connection"].items():
      road_section = dynamic_routing_utils._road_section_from_nodes(
          origin=o_node, destination=d_node)
      bpr_a_coefficient[road_section] = section_dict["a"]
      bpr_b_coefficient[road_section] = section_dict["b"]
      capacity[road_section] = section_dict["capacity"]
      free_flow_travel_time[road_section] = section_dict[
          "free_flow_travel_time"]
  node_position = {key: value["location"] for key, value in graph_dict.items()}
  return dynamic_routing_utils.Network(
      adjacency_list,
      node_position=node_position,
      bpr_a_coefficient=bpr_a_coefficient,
      bpr_b_coefficient=bpr_b_coefficient,
      capacity=capacity,
      free_flow_travel_time=free_flow_travel_time)


def create_series_parallel_network(num_network_in_series,
                                   time_step_length=1,
                                   capacity=1):
  i = 0
  origin = "A_0->B_0"
  graph_dict = {}
  while i < num_network_in_series:
    tt_up = random.random() + time_step_length
    tt_down = random.random() + time_step_length
    graph_dict.update({
        f"A_{i}": {
            "connection": {
                f"B_{i}": {
                    "a": 0,
                    "b": 1.0,
                    "capacity": capacity,
                    "free_flow_travel_time": time_step_length
                }
            },
            "location": [0 + 3 * i, 0]
        },
        f"B_{i}": {
            "connection": {
                f"C_{i}": {
                    "a": 1.0,
                    "b": 1.0,
                    "capacity": capacity,
                    "free_flow_travel_time": tt_up
                },
                f"D_{i}": {
                    "a": 1.0,
                    "b": 1.0,
                    "capacity": capacity,
                    "free_flow_travel_time": tt_down
                }
            },
            "location": [1 + 3 * i, 0]
        },
        f"C_{i}": {
            "connection": {
                f"A_{i+1}": {
                    "a": 0,
                    "b": 1.0,
                    "capacity": capacity,
                    "free_flow_travel_time": time_step_length
                }
            },
            "location": [2 + 3 * i, 1]
        },
        f"D_{i}": {
            "connection": {
                f"A_{i+1}": {
                    "a": 0,
                    "b": 1.0,
                    "capacity": capacity,
                    "free_flow_travel_time": time_step_length
                }
            },
            "location": [2 + 3 * i, -1]
        }
    })
    i += 1
  graph_dict[f"A_{i}"] = {
      "connection": {
          "END": {
              "a": 0,
              "b": 1.0,
              "capacity": capacity,
              "free_flow_travel_time": time_step_length
          }
      },
      "location": [0 + 3 * i, 0]
  }
  graph_dict["END"] = {"connection": {}, "location": [1 + 3 * i, 0]}
  time_horizon = int(3.0 * (num_network_in_series + 1) / time_step_length)
  destination = f"A_{i}->END"
  adjacency_list = {
      key: list(value["connection"].keys())
      for key, value in graph_dict.items()
  }
  bpr_a_coefficient = {}
  bpr_b_coefficient = {}
  capacity = {}
  free_flow_travel_time = {}
  for o_node, value_dict in graph_dict.items():
    for d_node, section_dict in value_dict["connection"].items():
      road_section = dynamic_routing_utils._road_section_from_nodes(
          origin=o_node, destination=d_node)
      bpr_a_coefficient[road_section] = section_dict["a"]
      bpr_b_coefficient[road_section] = section_dict["b"]
      capacity[road_section] = section_dict["capacity"]
      free_flow_travel_time[road_section] = section_dict[
          "free_flow_travel_time"]
  node_position = {key: value["location"] for key, value in graph_dict.items()}
  return dynamic_routing_utils.Network(
      adjacency_list,
      node_position=node_position,
      bpr_a_coefficient=bpr_a_coefficient,
      bpr_b_coefficient=bpr_b_coefficient,
      capacity=capacity,
      free_flow_travel_time=free_flow_travel_time
  ), origin, destination, time_horizon


def plot_network_n_player_game(g: dynamic_routing_utils.Network,
                               vehicle_locations=None):
  """Plot the network.

  Args:
    g: network to plot
    vehicle_locations: vehicle location
  """
  _, ax = plt.subplots()
  o_xs, o_ys, d_xs, d_ys = g.return_list_for_matplotlib_quiver()
  ax.quiver(
      o_xs,
      o_ys,
      np.subtract(d_xs, o_xs),
      np.subtract(d_ys, o_ys),
      color="b",
      angles="xy",
      scale_units="xy",
      scale=1)
  ax.set_xlim([
      np.min(np.concatenate((o_xs, d_xs))) - 0.5,
      np.max(np.concatenate((o_xs, d_xs))) + 0.5
  ])
  ax.set_ylim([
      np.min(np.concatenate((o_ys, d_ys))) - 0.5,
      np.max(np.concatenate((o_ys, d_ys))) + 0.5
  ])

  if vehicle_locations is not None:
    num_vehicle = len(vehicle_locations)
    dict_location = {}
    for vehicle_location in vehicle_locations:
      if vehicle_location not in dict_location:
        dict_location[vehicle_location] = 0.0
      dict_location[vehicle_location] += 0.3 / num_vehicle
    for point, width in dict_location.items():
      circle = plt.Circle(point, width, color="r")
      ax.add_patch(circle)


def plot_network_mean_field_game(g: dynamic_routing_utils.Network,
                                 distribution=None,
                                 scaling=1):
  """Plot the network.

  Args:
    g: network to plot
    distribution: the distribution.
    scaling: scaling factor. for plot rendering.
  """
  _, ax = plt.subplots()
  o_xs, o_ys, d_xs, d_ys = g.return_list_for_matplotlib_quiver()
  ax.quiver(
      o_xs,
      o_ys,
      np.subtract(d_xs, o_xs),
      np.subtract(d_ys, o_ys),
      color="b",
      angles="xy",
      scale_units="xy",
      scale=1)
  ax.set_xlim([
      np.min(np.concatenate((o_xs, d_xs))) - 0.5,
      np.max(np.concatenate((o_xs, d_xs))) + 0.5
  ])
  ax.set_ylim([
      np.min(np.concatenate((o_ys, d_ys))) - 0.5,
      np.max(np.concatenate((o_ys, d_ys))) + 0.5
  ])

  if distribution is not None:
    for x, prob_of_position in distribution.items():
      point = g.return_position_of_road_section(x)
      width = 0.3 * scaling * prob_of_position
      circle = plt.Circle(point, width, color="r")
      ax.add_patch(circle)


def evolve_n_player_simultaneous_game(game, policy, graph):
  state = game.new_initial_state()
  i = 0
  while not state.is_terminal():
    i += 1
    if state.is_chance_node():
      # Sample a chance event outcome.
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)
      state.apply_action(action)
    elif state.is_simultaneous_node():
      # Simultaneous node: sample actions for all players.
      chosen_actions = []
      for i in range(game.num_players()):
        legal_actions = state.legal_actions(i)
        state_policy = policy(state, i)
        assert len(legal_actions) == len(state_policy), (
            f"{legal_actions} not same length than {state_policy}")
        chosen_actions.append(
            random.choices(legal_actions,
                           [state_policy[a] for a in legal_actions])[0])
      state.apply_actions(chosen_actions)
    else:
      raise ValueError(
          "State should either be simultaneous node or change node.")
    plot_network_n_player_game(graph, [
        graph.return_position_of_road_section(x)
        for x in state.get_current_vehicle_locations()
    ])
  print(f"Travel times: {[-x for x in state.returns()]}")


def evolve_n_player_sequential_game(seq_game, policy, graph, debug=False):
  state = seq_game.new_initial_state()
  while not state.is_terminal():
    legal_actions = state.legal_actions()
    if state.is_chance_node():
      # Sample a chance event outcome.
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)
      if debug:
        print("------------ Change node ------------")
        print(
            (f"Possible chance actions: {outcomes_with_probs}, the one taken: "
             f"{action}."))
      state.apply_action(action)
    else:
      if debug:
        print("------------ Sequential action node ------------")
        print(state.information_state_tensor())
        print(state.observation_tensor())
        print(state.information_state_string())
      if policy is not None:
        state_policy = policy(state)
        vehicle_location = [
            s.replace("'", "")
            for s in str(state).split("[")[1].split("]")[0].split(", ")
        ]
        if debug:
          print((f"Policy for player {state.current_player()} at location "
                 f"{vehicle_location[state.current_player()]}: ") +
                str([(str(graph.get_road_section_from_action_id(k)) +
                      f"with probability {v}")
                     for k, v in state_policy.items()]))
        assert set(state_policy) == set(legal_actions)
        action = random.choices(legal_actions,
                                [state_policy[a] for a in legal_actions])
        assert len(action) == 1
        action = action[0]
      else:
        action = random.choice(legal_actions)
      state.apply_action(action)
      vehicle_location = [
          s.replace("'", "")
          for s in str(state).split("[")[1].split("]")[0].split(", ")
      ]
      if debug:
        print(vehicle_location)
      plot_network_n_player_game(
          graph,
          [graph.return_position_of_road_section(x) for x in vehicle_location])
  if debug:
    print(f"Travel times: {[-x for x in state.returns()]}")


def evolve_mean_field_game(mfg_game,
                           policy,
                           graph,
                           scaling=1,
                           frequency_printing=1):
  distribution_mfg = distribution_module.DistributionPolicy(mfg_game, policy)
  root_state = mfg_game.new_initial_state()
  listing_states = [root_state]

  # plot_network_mean_field_game(graph, {origin: 1})
  i = 0
  while not listing_states[0].is_terminal() and not all(
      state._vehicle_without_legal_action for state in listing_states):  # pylint:disable=protected-access
    assert abs(sum(map(distribution_mfg.value, listing_states)) - 1) < 1e-4, (
        f"{list(map(distribution_mfg.value, listing_states))}")
    new_listing_states = []
    list_of_state_seen = set()
    # In case chance node:
    if listing_states[0].current_player() == pyspiel.PlayerId.CHANCE:
      for mfg_state in listing_states:
        for action, _ in mfg_state.chance_outcomes():
          new_mfg_state = mfg_state.child(action)
          # Do not append twice the same file.
          if str(new_mfg_state) not in list_of_state_seen:
            new_listing_states.append(new_mfg_state)
          list_of_state_seen.add(str(new_mfg_state))
      current_distribution = {}
      for mfg_state in new_listing_states:
        location = mfg_state._vehicle_location  # pylint:disable=protected-access
        if location not in current_distribution:
          current_distribution[location] = 0
        current_distribution[location] += distribution_mfg.value(mfg_state)
      plot_network_mean_field_game(graph, current_distribution, scaling=scaling)

    # In case mean field node:
    elif listing_states[0].current_player() == pyspiel.PlayerId.MEAN_FIELD:
      for mfg_state in listing_states:
        dist_to_register = mfg_state.distribution_support()

        def get_probability_for_state(str_state):
          try:
            return distribution_mfg.value_str(str_state)
          except ValueError:
            return 0

        dist = [
            get_probability_for_state(str_state)
            for str_state in dist_to_register
        ]
        new_mfg_state = mfg_state.clone()
        new_mfg_state.update_distribution(dist)
        # Do not append twice the same file.
        if str(new_mfg_state) not in list_of_state_seen:
          new_listing_states.append(new_mfg_state)
        list_of_state_seen.add(str(new_mfg_state))

    # In case action node:
    else:
      assert (listing_states[0].current_player() ==
              pyspiel.PlayerId.DEFAULT_PLAYER_ID), "The player id should be 0"
      for mfg_state in listing_states:
        for action, _ in policy.action_probabilities(mfg_state).items():
          new_mfg_state = mfg_state.child(action)
          # Do not append twice the same file.
          if str(new_mfg_state) not in list_of_state_seen:
            new_listing_states.append(new_mfg_state)
          list_of_state_seen.add(str(new_mfg_state))
      current_distribution = {}
      for mfg_state in new_listing_states:
        location = mfg_state._vehicle_location  # pylint:disable=protected-access
        if location not in current_distribution:
          current_distribution[location] = 0
        current_distribution[location] += distribution_mfg.value(mfg_state)
      assert abs(sum(current_distribution.values()) - 1) < 1e-4, (
          f"{current_distribution}")
      i += 1
      if i % frequency_printing == 0:
        plot_network_mean_field_game(
            graph, current_distribution, scaling=scaling)
    listing_states = new_listing_states


def uniform_policy_n_player(seq_game):
  return policy_module.UniformRandomPolicy(seq_game)


def first_action_policy_n_player(seq_game):
  return policy_module.FirstActionPolicy(seq_game)


def ficticious_play(seq_game, number_of_iterations, compute_metrics=False):
  xfp_solver = fictitious_play.XFPSolver(seq_game)
  tick_time = time.time()
  for _ in range(number_of_iterations):
    xfp_solver.iteration()
  timing = time.time() - tick_time
  # print('done')
  # average_policies = xfp_solver.average_policy_tables()
  tabular_policy = policy_module.TabularPolicy(seq_game)
  if compute_metrics:
    nash_conv = exploitability.nash_conv(seq_game, xfp_solver.average_policy())
    average_policy_values = expected_game_score.policy_value(
        seq_game.new_initial_state(), [tabular_policy])
    return timing, tabular_policy, nash_conv, average_policy_values
  return timing, tabular_policy


def counterfactual_regret_minimization(seq_game,
                                       number_of_iterations,
                                       compute_metrics=False):
  # freq_iteration_printing = number_of_iterations // 10
  cfr_solver = cfr.CFRSolver(seq_game)
  tick_time = time.time()
  # print("CFRSolver initialized.")
  for _ in range(number_of_iterations):
    cfr_solver.evaluate_and_update_policy()
    # if i % freq_iteration_printing == 0:
    #   print(f"Iteration {i}")
  timing = time.time() - tick_time
  # print("Finish.")
  if compute_metrics:
    nash_conv = exploitability.nash_conv(seq_game, cfr_solver.average_policy())
    return timing, cfr_solver.average_policy(), nash_conv
  return timing, cfr_solver.average_policy()


def external_sampling_monte_carlo_counterfactual_regret_minimization(
    seq_game, number_of_iterations, compute_metrics=False):
  cfr_solver = external_mccfr.ExternalSamplingSolver(
      seq_game, external_mccfr.AverageType.SIMPLE)
  tick_time = time.time()
  # print("CFRSolver initialized.")
  for _ in range(number_of_iterations):
    cfr_solver.iteration()
  timing = time.time() - tick_time
  # print("Finish.")
  if compute_metrics:
    nash_conv = exploitability.nash_conv(seq_game, cfr_solver.average_policy())
    return timing, cfr_solver.average_policy(), nash_conv
  return timing, cfr_solver.average_policy()


class NFSPPolicies(policy_module.Policy):
  """Joint policy to be evaluated."""

  def __init__(self, env, nfsp_policies, mode):
    game = env.game
    num_players = env.num_players
    player_ids = list(range(num_players))
    super().__init__(game, player_ids)
    self._policies = nfsp_policies
    self._mode = mode
    self._obs = {
        "info_state": [None] * num_players,
        "legal_actions": [None] * num_players
    }

  def action_probabilities(self, state, player_id=None):
    del player_id
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)

    self._obs["current_player"] = cur_player
    self._obs["info_state"][cur_player] = (
        state.information_state_tensor(cur_player))
    self._obs["legal_actions"][cur_player] = legal_actions

    info_state = rl_environment.TimeStep(
        observations=self._obs, rewards=None, discounts=None, step_type=None)

    with self._policies[cur_player].temp_mode_as(self._mode):
      p = self._policies[cur_player].step(info_state, is_evaluation=True).probs
    prob_dict = {action: p[action] for action in legal_actions}
    return prob_dict


def neural_ficticious_self_play(seq_game,
                                num_epoch,
                                sess,
                                compute_metrics=False):
  env = rl_environment.Environment(seq_game)
  # Parameters from the game.
  num_players = env.num_players
  num_actions = env.action_spec()["num_actions"]
  info_state_size = env.observation_spec()["info_state"][0]

  # Parameters for the algorithm.
  hidden_layers_sizes = [int(l) for l in [128]]

  kwargs = {
      "replay_buffer_capacity": int(2e5),
      "reservoir_buffer_capacity": int(2e6),
      "min_buffer_size_to_learn": 1000,
      "anticipatory_param": 0.1,
      "batch_size": 128,
      "learn_every": 64,
      "rl_learning_rate": 0.01,
      "sl_learning_rate": 0.01,
      "optimizer_str": "sgd",
      "loss_str": "mse",
      "update_target_network_every": 19200,
      "discount_factor": 1.0,
      "epsilon_decay_duration": int(20e6),
      "epsilon_start": 0.06,
      "epsilon_end": 0.001,
  }

  # freq_epoch_printing = num_epoch // 10
  agents = [
      nfsp.NFSP(sess, idx, info_state_size, num_actions, hidden_layers_sizes,
                **kwargs) for idx in range(num_players)
  ]
  joint_avg_policy = NFSPPolicies(env, agents, nfsp.MODE.average_policy)

  sess.run(tf.global_variables_initializer())
  # print("TF initialized.")
  tick_time = time.time()
  for _ in range(num_epoch):
    # if ep % freq_epoch_printing == 0:
    #   print(f"Iteration {ep}")
    time_step = env.reset()
    while not time_step.last():
      player_id = time_step.observations["current_player"]
      agent_output = agents[player_id].step(time_step)
      action_list = [agent_output.action]
      time_step = env.step(action_list)

    # Episode is over, step all agents with final info state.
    for agent in agents:
      agent.step(time_step)
  timing = time.time() - tick_time
  # print("Finish.")
  if compute_metrics:
    tabular_policy = joint_avg_policy.TabularPolicy(seq_game)
    average_policy_values = expected_game_score.policy_value(
        seq_game.new_initial_state(), [tabular_policy])
    nash_conv = exploitability.nash_conv(env.game, joint_avg_policy)
    return timing, joint_avg_policy, average_policy_values, nash_conv
  return timing, joint_avg_policy


def mean_field_uniform_policy(mfg_game,
                              number_of_iterations,
                              compute_metrics=False):
  del number_of_iterations
  uniform_policy = policy_module.UniformRandomPolicy(mfg_game)
  if compute_metrics:
    distribution_mfg = distribution_module.DistributionPolicy(
        mfg_game, uniform_policy)
    policy_value_ = policy_value.PolicyValue(mfg_game, distribution_mfg,
                                             uniform_policy).value(
                                                 mfg_game.new_initial_state())
    return uniform_policy, policy_value_
  return uniform_policy


def mean_field_fictitious_play(mfg_game,
                               number_of_iterations,
                               compute_metrics=False):
  fp = mean_field_fictitious_play_module.FictitiousPlay(mfg_game)
  tick_time = time.time()
  for _ in range(number_of_iterations):
    fp.iteration()
  timing = time.time() - tick_time
  fp_policy = fp.get_policy()
  # print('learning done')
  if compute_metrics:
    distribution_mfg = distribution_module.DistributionPolicy(
        mfg_game, fp_policy)
    # print('distribution done')
    policy_value_ = policy_value.PolicyValue(mfg_game, distribution_mfg,
                                             fp_policy).value(
                                                 mfg_game.new_initial_state())
    nash_conv_fp = nash_conv_module.NashConv(mfg_game, fp_policy)
    return timing, fp_policy, nash_conv_fp, policy_value_
  return timing, fp_policy


def online_mirror_descent(mfg_game,
                          number_of_iterations,
                          compute_metrics=False,
                          return_policy=False,
                          md_p=None):
  md = md_p if md_p else mirror_descent.MirrorDescent(mfg_game)
  tick_time = time.time()
  for _ in range(number_of_iterations):
    md.iteration()
  timing = time.time() - tick_time
  md_policy = md.get_policy()
  if compute_metrics:
    distribution_mfg = distribution_module.DistributionPolicy(
        mfg_game, md_policy)
    # print('distribution done')
    policy_value_ = policy_value.PolicyValue(mfg_game, distribution_mfg,
                                             md_policy).value(
                                                 mfg_game.new_initial_state())
    nash_conv_md = nash_conv_module.NashConv(mfg_game, md_policy)
    if return_policy:
      return timing, md_policy, nash_conv_md, policy_value_, md
    return timing, md_policy, nash_conv_md, policy_value_
  return timing, md_policy


class RandomPolicyDeviation:

  def __init__(self):
    self.policy_deviation = {}

  def get_policy_deviation(self, state, player_id):
    key = (str(state), player_id)
    if key not in self.policy_deviation:
      assert player_id == state.current_player()
      action_probability = [random.random() for a in state.legal_actions()]
      self.policy_deviation[key] = [
          x / sum(action_probability) for x in action_probability
      ]
    return self.policy_deviation[key]


def get_results_n_player_sequential_game(seq_game, policy):
  state = seq_game.new_initial_state()
  while not state.is_terminal():
    legal_actions = state.legal_actions()
    if state.is_chance_node():
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)
    else:
      state_policy = policy(state)
      assert set(state_policy) == set(legal_actions)
      action = random.choices(legal_actions,
                              [state_policy[a] for a in legal_actions])
      assert len(action) == 1
      action = action[0]
    state.apply_action(action)
  return state.returns()


def get_list_results_n_player_game(seq_game, policy, num_sample=10):
  return [
      get_results_n_player_sequential_game(seq_game, policy)
      for _ in range(num_sample)
  ]


def get_average_results_n_player_game(seq_game, policy, num_sample=10):
  result_array = get_list_results_n_player_game(seq_game, policy, num_sample)
  return sum([sum(i) / len(i) for i in zip(*result_array)]) / len(result_array)


def get_results_n_player_simultaneous_game(game, policy):
  state = game.new_initial_state()
  i = 0
  while not state.is_terminal():
    i += 1
    if state.is_chance_node():
      # Sample a chance event outcome.
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)
      state.apply_action(action)
    elif state.is_simultaneous_node():
      # Simultaneous node: sample actions for all players.
      chosen_actions = []
      for i in range(game.num_players()):
        legal_actions = state.legal_actions(i)
        state_policy = policy(state, player_id=i)
        assert abs(sum([state_policy[a] for a in legal_actions]) - 1) < 1e-4
        chosen_actions.append(
            random.choices(legal_actions,
                           [state_policy[a] for a in legal_actions])[0])
      state.apply_actions(chosen_actions)
    else:
      raise ValueError(
          "State should either be simultaneous node or change node.")
  return state.returns()


def get_list_results_n_player_simulataneous_game(game, policy, num_sample=10):
  return [
      get_results_n_player_simultaneous_game(game, policy)
      for _ in range(num_sample)
  ]


def get_expected_value(seq_game, policy, num_sample, player=0):
  results = get_list_results_n_player_game(
      seq_game, policy, num_sample=num_sample)
  expected_value = sum(x[player] for x in results) / num_sample
  # num_vehicle = len(results[0])
  # error_bar = abs(sum([x[1] for x in results]) - sum(
  # [x[2] for x in results])) / num_sample_trajectories
  # expected_value_policy = sum(sum(x[i] for x in results) for i in range(
  # 1, BRAESS_NUM_VEHICLES)) / ((BRAESS_NUM_VEHICLES-1)*num_sample_trajectories)
  return expected_value


def compute_regret_policy(game,
                          policy,
                          num_random_policy_tested=10,
                          num_sample=100):
  time_tick = time.time()
  expected_value_policy = get_expected_value(game, policy, num_sample)
  worse_regret = 0
  for _ in range(num_random_policy_tested):
    noisy_n_policy = noisy_policy.NoisyPolicy(policy, player_id=0, alpha=1)
    expected_value_noise = get_expected_value(
        game, noisy_n_policy, num_sample, player=0)
    approximate_regret = expected_value_noise - expected_value_policy
    worse_regret = max(worse_regret, approximate_regret)
  return worse_regret, time.time() - time_tick


def get_expected_value_sim_game(game, policy, num_sample, player=0):
  results = get_list_results_n_player_simulataneous_game(
      game, policy, num_sample=num_sample)
  assert len(results) == num_sample
  expected_value = sum(x[player] for x in results) / num_sample
  # num_vehicle = len(results[0])
  # error_bar = abs(sum([x[1] for x in results]) - sum(
  # [x[2] for x in results])) / num_sample_trajectories
  # expected_value_policy = sum(sum(x[i] for x in results) for i in range(
  # 1, BRAESS_NUM_VEHICLES)) / ((BRAESS_NUM_VEHICLES-1)*num_sample_trajectories)
  return expected_value


def compute_regret_policy_random_noise_sim_game(game,
                                                policy,
                                                num_random_policy_tested=10,
                                                num_sample=100):
  time_tick = time.time()
  expected_value_policy = get_expected_value_sim_game(game, policy, num_sample)
  worse_regret = 0
  for _ in range(num_random_policy_tested):
    noisy_n_policy = noisy_policy.NoisyPolicy(policy, player_id=0, alpha=1)
    expected_value_noise = get_expected_value_sim_game(
        game, noisy_n_policy, num_sample, player=0)
    approximate_regret = expected_value_noise - expected_value_policy
    worse_regret = max(worse_regret, approximate_regret)
  return worse_regret, time.time() - time_tick


class PurePolicyResponse(policy_module.Policy):

  def __init__(self, game, policy, player_id):
    self.game = game
    self.player_id = player_id
    self.policy = policy

  def pure_action(self, state):
    raise NotImplementedError()

  def action_probabilities(self, state, player_id=None):
    assert player_id is not None
    if player_id == self.player_id:
      legal_actions = state.legal_actions(self.player_id)
      if not legal_actions:
        return {0: 1.0}
      if len(legal_actions) == 1:
        return {legal_actions[0]: 1.0}
      answer = {action: 0.0 for action in legal_actions}
      pure_a = self.pure_action(state)
      assert pure_a in answer
      answer[pure_a] = 1.0
      return answer
    return self.policy.action_probabilities(state, player_id)


class PathBCEResponse(PurePolicyResponse):

  def pure_action(self, state):
    location = state.get_current_vehicle_locations()[self.player_id].split(
        "->")[1]
    if location == "B":
      return state.get_game().network.get_action_id_from_movement("B", "C")
    if location == "C":
      return state.get_game().network.get_action_id_from_movement("C", "E")
    return 0


class PathBCDEResponse(PurePolicyResponse):

  def pure_action(self, state):
    location = state.get_current_vehicle_locations()[self.player_id].split(
        "->")[1]
    if location == "B":
      return state.get_game().network.get_action_id_from_movement("B", "C")
    if location == "C":
      return state.get_game().network.get_action_id_from_movement("C", "D")
    return 0


class PathBDEResponse(PurePolicyResponse):

  def pure_action(self, state):
    location = state.get_current_vehicle_locations()[self.player_id].split(
        "->")[1]
    if location == "B":
      return state.get_game().network.get_action_id_from_movement("B", "D")
    return 0


def compute_regret_policy_against_pure_policy_sim_game(game,
                                                       policy,
                                                       compute_true_value=False,
                                                       num_sample=100):
  time_tick = time.time()
  if compute_true_value:
    expected_value_policy = expected_game_score.policy_value(
        game.new_initial_state(), policy)[0]
  else:
    expected_value_policy = get_expected_value_sim_game(game, policy,
                                                        num_sample)
  worse_regret = 0
  policies = [
      PathBCEResponse(game, policy, 0),
      PathBCDEResponse(game, policy, 0),
      PathBDEResponse(game, policy, 0)
  ]
  for deviation_policy in policies:
    if compute_true_value:
      expected_value_noise = expected_game_score.policy_value(
          game.new_initial_state(), deviation_policy)[0]
    else:
      expected_value_noise = get_expected_value_sim_game(
          game, deviation_policy, num_sample, player=0)
    approximate_regret = expected_value_noise - expected_value_policy
    worse_regret = max(worse_regret, approximate_regret)
  return worse_regret, time.time() - time_tick


def online_mirror_descent_sioux_falls(mfg_game,
                                      number_of_iterations,
                                      md_p=None):
  nash_conv_dict = {}
  md = md_p if md_p else mirror_descent.MirrorDescent(mfg_game)
  tick_time = time.time()
  for i in range(number_of_iterations):
    md.iteration()
    md_policy = md.get_policy()
    nash_conv_md = nash_conv_module.NashConv(mfg_game, md_policy)
    nash_conv_dict[i] = nash_conv_md.nash_conv()
    print((f"Iteration {i}, Nash conv: {nash_conv_md.nash_conv()}, "
           "time: {time.time() - tick_time}"))
  timing = time.time() - tick_time
  md_policy = md.get_policy()
  distribution_mfg = distribution_module.DistributionPolicy(mfg_game, md_policy)
  policy_value_ = policy_value.PolicyValue(
      mfg_game, distribution_mfg, md_policy).value(mfg_game.new_initial_state())
  nash_conv_md = nash_conv_module.NashConv(mfg_game, md_policy)
  return timing, md_policy, nash_conv_md, policy_value_, md, nash_conv_dict



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAX implementation of LOLA and LOLA-DiCE (Foerster et al. 2018).

The DiCE implementation is also based on the pytorch implementation from
https://github.com/alexis-jacq/LOLA_DiCE by Alexis David Jacq.

Both algorithm implementations, LOLA and LOLA-DiCE, currently support only
two-player simultaneous move games and assume access to the opponent's
actions (the observation field in the time step must contain a key
'actions' with the opponent's actions).
"""

# pylint: disable=g-importing-member
# pylint: disable=g-bare-generic

from copy import deepcopy
from functools import partial
import typing

import chex
import distrax
import haiku as hk
import jax
from jax import grad
from jax import vmap
import jax.numpy as jnp
import numpy as np
import optax
import rlax

from open_spiel.python import rl_agent
from open_spiel.python import rl_environment
from open_spiel.python.rl_environment import TimeStep


@chex.dataclass
class TransitionBatch:  # pylint: disable=too-few-public-methods
  """A transition batch is a collection of transitions.

  Each item in the batch is a numpy array.
  """

  info_state: np.ndarray
  action: np.ndarray
  reward: np.ndarray
  discount: np.ndarray = None
  terminal: np.ndarray = None
  legal_actions_mask: np.ndarray = None
  values: np.ndarray = None


@chex.dataclass
class TrainState:  # pylint: disable=too-few-public-methods
  """TrainState class.

  The training state contains the parameters and optimizer states of the
  policy and critic networks for each agent. The parameters are stored in a
  dictionary with the agent id as key.
  """

  policy_params: typing.Dict[typing.Any, hk.Params]
  policy_opt_states: typing.Dict[typing.Any, optax.OptState]
  critic_params: typing.Dict[typing.Any, hk.Params]
  critic_opt_states: typing.Dict[typing.Any, optax.OptState]


# A function that takes the current train state and a transition batch and
# returns the new train state and a dictionary of metrics.
UpdateFn = typing.Callable[
    [TrainState, TransitionBatch], typing.Tuple[TrainState, typing.Dict]
]


def get_minibatches(
    batch: TransitionBatch, num_minibatches: int
) -> typing.Iterator[TransitionBatch]:
  """Yields an iterator over minibatches of the given batch.

  Args:
      batch: A transition batch.
      num_minibatches: The number of minibatches to return.

  Yields:
      An iterator over minibatches of the given batch.
  """

  def get_minibatch(x, start, end):
    return x[:, start:end] if len(x.shape) > 2 else x

  for i in range(num_minibatches):
    start, end = i * (batch.reward.shape[1] // num_minibatches), (i + 1) * (
        batch.reward.shape[1] // num_minibatches
    )
    mini_batch = jax.tree_util.tree_map(
        partial(get_minibatch, start=start, end=end), batch
    )
    yield mini_batch


def get_critic_update_fn(
    agent_id: int,
    critic_network: hk.Transformed,
    optimizer: optax.TransformUpdateFn,
    num_minibatches: int = 8,
    gamma: float = 0.99,
) -> UpdateFn:
  """Returns the update function for the critic parameters.

  Args:
      agent_id: The id of the agent that will be updated.
      critic_network: A transformed haiku function.
      optimizer: Optimizer update function.
      num_minibatches: the number of minibatches.
      gamma: the discount factor.

  Returns:
      An update function that takes the current train state together with a
      transition batch and returns the new train state and a dictionary of
      metrics.
  """

  def loss_fn(params, batch: TransitionBatch):
    info_states, rewards = batch.info_state[agent_id], batch.reward[agent_id]
    discounts = jnp.ones_like(rewards) * gamma
    values = critic_network.apply(params, info_states).squeeze()
    v_t = values[:, :-1].reshape(-1)
    v_tp1 = values[:, 1:].reshape(-1)
    r_t = rewards[:, :-1].reshape(-1)
    d_t = discounts[:, 1:].reshape(-1)
    td_error = jax.lax.stop_gradient(r_t + d_t * v_tp1) - v_t
    return jnp.mean(td_error**2)

  def update(train_state: TrainState, batch: TransitionBatch):
    """The critic update function.

    Updates the critic parameters of the train state with the given
    transition batch.
    
    Args:
        train_state: The current train state.
        batch: A transition batch.

    Returns:
        The updated train state with the new critic params and a dictionary
        with the critic loss
    """
    losses = []
    critic_params = train_state.critic_params[agent_id]
    opt_state = train_state.critic_opt_states[agent_id]
    for mini_batch in get_minibatches(batch, num_minibatches):
      loss, grads = jax.value_and_grad(loss_fn)(critic_params, mini_batch)
      updates, opt_state = optimizer(grads, opt_state)
      critic_params = optax.apply_updates(critic_params, updates)
      losses.append(loss)
    train_state = deepcopy(train_state)
    state = TrainState(
        policy_params=train_state.policy_params,
        policy_opt_states=train_state.policy_opt_states,
        critic_params={**train_state.critic_params, agent_id: critic_params},
        critic_opt_states={
            **train_state.critic_opt_states,
            agent_id: opt_state,
        },
    )
    return state, {'loss': jnp.mean(jnp.array(losses))}

  return update


def get_dice_update_fn(
    agent_id: int,
    rng: hk.PRNGSequence,
    policy_network: hk.Transformed,
    critic_network: hk.Transformed,
    optimizer: optax.TransformUpdateFn,
    opp_pi_lr: float,
    env: rl_environment.Environment,
    n_lookaheads: int = 1,
    gamma: float = 0.99,
):
  """Get the DiCE update function."""
  def magic_box(x):
    return jnp.exp(x - jax.lax.stop_gradient(x))

  @jax.jit
  @partial(jax.vmap, in_axes=(None, 0, 0))
  def get_action(params, s, rng_key):
    pi = policy_network.apply(params, s)
    action = pi.sample(seed=rng_key)
    return action

  def rollout(params, other_params):
    states, rewards, actions = [], [], []
    step = env.reset()
    batch_size = (
        step.observations['batch_size']
        if 'batch_size' in step.observations
        else 1
    )
    while not step.last():
      obs = step.observations
      s_1, s_2 = jnp.array(obs['info_state'][0]), jnp.array(
          obs['info_state'][1]
      )
      if batch_size == 1:
        s_1, s_2 = s_1[None, :], s_2[None, :]
      a_1 = get_action(params, s_1, jax.random.split(next(rng), num=batch_size))
      a_2 = get_action(
          other_params, s_2, jax.random.split(next(rng), num=batch_size)
      )
      a = jnp.stack([a_1, a_2], axis=1)
      step = env.step(a.squeeze())
      r_1, r_2 = jnp.array(step.rewards[0]), jnp.array(step.rewards[1])
      if batch_size == 1:
        r_1, r_2 = r_1[None], r_2[None]
      actions.append(a.T)
      states.append(jnp.stack([s_1, s_2], axis=0))
      rewards.append(jnp.stack([r_1, r_2], axis=0))
    return {
        'states': jnp.stack(states, axis=2),
        'rewards': jnp.stack(rewards, axis=2),
        'actions': jnp.stack(actions, axis=2),
    }

  def dice_correction(train_state: TrainState):
    """Computes the dice update for the given train state.

    Args:
        train_state: The current train state.

    Returns:
        The updated train state with the new policy params and metrics dict.
    """

    @jax.jit
    def dice_objective(params, other_params, states, actions, rewards, values):
      self_logprobs = vmap(
          vmap(lambda s, a: policy_network.apply(params, s).log_prob(a))
      )(states[0], actions[0])
      other_logprobs = vmap(
          vmap(lambda s, a: policy_network.apply(other_params, s).log_prob(a))
      )(states[1], actions[1])
      # apply discount:
      cum_discount = jnp.cumprod(gamma * jnp.ones_like(rewards), axis=1) / gamma
      discounted_rewards = rewards * cum_discount
      discounted_values = values.squeeze() * cum_discount

      # stochastics nodes involved in rewards dependencies:
      dependencies = jnp.cumsum(self_logprobs + other_logprobs, axis=1)
      # logprob of each stochastic nodes:
      stochastic_nodes = self_logprobs + other_logprobs
      # dice objective:
      dice_objective = jnp.mean(
          jnp.sum(magic_box(dependencies) * discounted_rewards, axis=1)
      )
      baseline_term = jnp.mean(
          jnp.sum((1 - magic_box(stochastic_nodes)) * discounted_values, axis=1)
      )
      dice_objective = dice_objective + baseline_term
      return -dice_objective  # want to minimize -objective

    def outer_update(params, opp_params, agent_id, opp_id):
      other_theta = opp_params
      for _ in range(n_lookaheads):
        trajectories = rollout(other_theta, params)
        other_grad = jax.grad(dice_objective)(
            other_theta,
            other_params=params,
            states=trajectories['states'],
            actions=trajectories['actions'],
            rewards=trajectories['rewards'][0],
            values=critic_network.apply(
                train_state.critic_params[opp_id], trajectories['states'][0]
            ),
        )
        # Update the other player's policy:
        other_theta = jax.tree_util.tree_map(
            lambda param, grad: param - opp_pi_lr * grad,
            other_theta,
            other_grad,
        )

      trajectories = rollout(params, other_theta)
      values = critic_network.apply(
          train_state.critic_params[agent_id], trajectories['states'][0]
      )
      loss = dice_objective(
          params=params,
          other_params=other_theta,
          states=trajectories['states'],
          actions=trajectories['actions'],
          rewards=trajectories['rewards'][0],
          values=values,
      )
      return loss, {'loss': loss}

    opp = 1 - agent_id
    grads, metrics = grad(outer_update, has_aux=True)(
        train_state.policy_params[agent_id],
        opp_params=train_state.policy_params[opp],
        agent_id=agent_id,
        opp_id=opp,
    )
    return grads, metrics

  def update(
      train_state: TrainState, batch: TransitionBatch
  ) -> typing.Tuple[TrainState, typing.Dict]:
    """Updates the policy parameters in train_state.

    If lola_weight > 0, the correction term according to Foerster et al. will be
    applied.

    Args:
        train_state: the agent's train state.
        batch: a transition batch

    Returns:
        A tuple (new_train_state, metrics)
    """
    del batch
    grads, metrics = dice_correction(train_state)
    updates, opt_state = optimizer(
        grads, train_state.policy_opt_states[agent_id]
    )
    policy_params = optax.apply_updates(
        train_state.policy_params[agent_id], updates
    )
    train_state = TrainState(
        policy_params={**train_state.policy_params, agent_id: policy_params},
        policy_opt_states={
            **train_state.policy_opt_states,
            agent_id: opt_state,
        },
        critic_params=deepcopy(train_state.critic_params),
        critic_opt_states=deepcopy(train_state.critic_opt_states),
    )
    return train_state, metrics

  return update


def get_lola_update_fn(
    agent_id: int,
    policy_network: hk.Transformed,
    optimizer: optax.TransformUpdateFn,
    pi_lr: float,
    gamma: float = 0.99,
    lola_weight: float = 1.0,
) -> UpdateFn:
  """Get the LOLA update function.

  Returns a function that updates the policy parameters using the LOLA
  correction formula.
  
  Args:
      agent_id: the agent's id
      policy_network: A haiku transformed policy network.
      optimizer: An optax optimizer.
      pi_lr: Policy learning rate.
      gamma: Discount factor.
      lola_weight: The LOLA correction weight to scale the correction term.

  Returns:
      A UpdateFn function that updates the policy parameters.
  """

  def flat_params(
      params,
  ) -> typing.Tuple[
      typing.Dict[str, jnp.ndarray], typing.Dict[typing.Any, typing.Callable]
  ]:
    """Flattens the policy parameters.
    
    Flattens the parameters of the policy network into a single vector and
    returns the unravel function.
    
    Args:
        params: The policy parameters.

    Returns:
        A tuple (flat_params, unravel_fn)
    """
    flat_param_dict = {
        agent_id: jax.flatten_util.ravel_pytree(p)
        for agent_id, p in params.items()
    }

    params = dict((k, flat_param_dict[k][0]) for k in flat_param_dict)
    unravel_fns = dict((k, flat_param_dict[k][1]) for k in flat_param_dict)
    return params, unravel_fns

  def lola_correction(
      train_state: TrainState, batch: TransitionBatch
  ) -> hk.Params:
    """Computes the LOLA correction term.

    Args:
        train_state: The agent's current train state.
        batch: A transition batch.

    Returns:
        The LOLA correction term.
    """
    a_t, o_t, r_t, values = (
        batch.action,
        batch.info_state,
        batch.reward,
        batch.values,
    )
    params, unravel_fns = flat_params(train_state.policy_params)

    compute_returns = partial(rlax.lambda_returns, lambda_=0.0)
    g_t = vmap(vmap(compute_returns))(
        r_t=r_t, v_t=values, discount_t=jnp.full_like(r_t, gamma)
    )
    g_t = (g_t - g_t.mean()) / (g_t.std() + 1e-8)

    def log_pi(params, i, a_t, o_t):
      return policy_network.apply(unravel_fns[i](params), o_t).log_prob(a_t)

    opp_id = 1 - agent_id

    def cross_term(a_t, o_t, r_t):
      """Computes the second order correction term of the LOLA update.

      Args:
          a_t: actions of both players
          o_t: observations of both players
          r_t: rewards of both players

      Returns:
          The second order correction term.
      """
      grad_log_pi = vmap(jax.value_and_grad(log_pi), in_axes=(None, None, 0, 0))
      log_probs, grads = grad_log_pi(
          params[agent_id], agent_id, a_t[agent_id], o_t[agent_id]
      )
      opp_logrpobs, opp_grads = grad_log_pi(
          params[opp_id], opp_id, a_t[opp_id], o_t[opp_id]
      )
      grads = grads.cumsum(axis=0)
      opp_grads = opp_grads.cumsum(axis=0)
      log_probs = log_probs.cumsum(axis=0)
      opp_logrpobs = opp_logrpobs.cumsum(axis=0)
      cross_term = 0.0
      for t in range(0, len(a_t[agent_id])):
        discounted_reward = r_t[opp_id, t] * jnp.power(gamma, t)
        cross_term += (
            discounted_reward
            * jnp.outer(grads[t], opp_grads[t])
            * jnp.exp(log_probs[t] + opp_logrpobs[t])
        )
      return cross_term  # * jnp.exp(log_probs.sum() + opp_logrpobs.sum())

    def policy_gradient(a_t, o_t, g_t):
      grad_log_pi = vmap(grad(log_pi), in_axes=(None, None, 0, 0))
      opp_grads = grad_log_pi(params[opp_id], opp_id, a_t[opp_id], o_t[opp_id])
      pg = g_t[agent_id] @ opp_grads
      return pg

    cross = vmap(cross_term, in_axes=(1, 1, 1))(a_t, o_t, r_t).mean(axis=0)
    pg = vmap(policy_gradient, in_axes=(1, 1, 1))(a_t, o_t, g_t).mean(axis=0)
    correction = -pi_lr * (pg @ cross)
    return unravel_fns[agent_id](correction)

  def policy_loss(params, agent_id, batch):
    """Computes the policy gradient loss.

    Args:
        params: The policy parameters.
        agent_id: The agent's id.
        batch: A transition batch.

    Returns:
        The policy gradient loss.
    """
    a_t, o_t, r_t, values = (
        batch.action[agent_id],
        batch.info_state[agent_id],
        batch.reward[agent_id],
        batch.values[agent_id],
    )
    logits_t = vmap(vmap(lambda s: policy_network.apply(params, s).logits))(o_t)
    discount = jnp.full(r_t.shape, gamma)
    returns = vmap(rlax.lambda_returns)(
        r_t=r_t,
        v_t=values,
        discount_t=discount,
        lambda_=jnp.ones_like(discount),
    )
    adv_t = returns - values
    loss = vmap(rlax.policy_gradient_loss)(
        logits_t=logits_t, a_t=a_t, adv_t=adv_t, w_t=jnp.ones_like(adv_t)
    )
    return loss.mean()

  def update(
      train_state: TrainState, batch: TransitionBatch
  ) -> typing.Tuple[TrainState, typing.Dict]:
    """Updates the policy parameters in train_state.

    If lola_weight > 0, the correction term by Foerster et al. will be applied.

    Args:
        train_state: the agent's train state.
        batch: a transition batch

    Returns:
        A tuple (new_train_state, metrics)
    """
    loss, policy_grads = jax.value_and_grad(policy_loss)(
        train_state.policy_params[agent_id], agent_id, batch
    )
    correction = lola_correction(train_state, batch)
    policy_grads = jax.tree_util.tree_map(
        lambda grad, corr: grad - lola_weight * corr, policy_grads, correction
    )
    updates, opt_state = optimizer(
        policy_grads, train_state.policy_opt_states[agent_id]
    )
    policy_params = optax.apply_updates(
        train_state.policy_params[agent_id], updates
    )
    train_state = TrainState(
        policy_params={**train_state.policy_params, agent_id: policy_params},
        policy_opt_states={
            **train_state.policy_opt_states,
            agent_id: opt_state,
        },
        critic_params=deepcopy(train_state.critic_params),
        critic_opt_states=deepcopy(train_state.critic_opt_states),
    )
    return train_state, {'loss': loss}

  return update


def get_opponent_update_fn(
    agent_id: int,
    policy_network: hk.Transformed,
    optimizer: optax.TransformUpdateFn,
    num_minibatches: int = 1,
) -> UpdateFn:
  """Get the opponent update function."""
  def loss_fn(params, batch: TransitionBatch):
    def loss(p, states, actions):
      log_prob = policy_network.apply(p, states).log_prob(actions)
      return log_prob

    log_probs = vmap(vmap(loss, in_axes=(None, 0, 0)), in_axes=(None, 0, 0))(
        params, batch.info_state[agent_id], batch.action[agent_id]
    )
    return -log_probs.sum(axis=-1).mean()

  def update(
      train_state: TrainState, batch: TransitionBatch
  ) -> typing.Tuple[TrainState, typing.Dict]:
    policy_params = train_state.policy_params[agent_id]
    opt_state = train_state.policy_opt_states[agent_id]
    loss = 0
    for mini_batch in get_minibatches(batch, num_minibatches):
      loss, policy_grads = jax.value_and_grad(loss_fn)(
          policy_params, mini_batch
      )
      updates, opt_state = optimizer(policy_grads, opt_state)
      policy_params = optax.apply_updates(
          train_state.policy_params[agent_id], updates
      )

    train_state = TrainState(
        policy_params={**train_state.policy_params, agent_id: policy_params},
        policy_opt_states={
            **train_state.policy_opt_states,
            agent_id: opt_state,
        },
        critic_params=deepcopy(train_state.critic_params),
        critic_opt_states=deepcopy(train_state.critic_opt_states),
    )
    return train_state, {'loss': loss}

  return update


class OpponentShapingAgent(rl_agent.AbstractAgent):
  """Opponent Shaping Agent.
  
  This agent uses either LOLA or LOLA-DiCE to influence the parameter updates
  of the opponent policies.
  """

  def __init__(
      self,
      player_id: int,
      opponent_ids: typing.List[int],
      info_state_size: chex.Shape,
      num_actions: int,
      policy: hk.Transformed,
      critic: hk.Transformed,
      batch_size: int = 16,
      critic_learning_rate: typing.Union[float, optax.Schedule] = 0.01,
      pi_learning_rate: typing.Union[float, optax.Schedule] = 0.001,
      opp_policy_learning_rate: typing.Union[float, optax.Schedule] = 0.001,
      opponent_model_learning_rate: typing.Union[float, optax.Schedule] = 0.001,
      clip_grad_norm: float = 0.5,
      policy_update_interval: int = 8,
      discount: float = 0.99,
      critic_discount: float = 0.99,
      seed: jax.random.PRNGKey = 42,
      fit_opponent_model=True,
      correction_type: str = 'dice',
      use_jit: bool = False,
      n_lookaheads: int = 1,
      num_critic_mini_batches: int = 1,
      num_opponent_updates: int = 1,
      env: typing.Optional[rl_environment.Environment] = None,
  ):
    self.player_id = player_id
    self._num_actions = num_actions
    self._batch_size = batch_size
    self._policy_update_interval = policy_update_interval
    self._discount = discount
    self._num_opponent_updates = num_opponent_updates
    self._num_mini_batches = num_critic_mini_batches
    self._prev_time_step = None
    self._prev_action = None
    self._data = []
    self._metrics = []
    self._fit_opponent_model = fit_opponent_model
    self._opponent_ids = opponent_ids
    self._rng = hk.PRNGSequence(seed)

    # Step counters
    self._step_counter = 0
    self._episode_counter = 0
    self._num_learn_steps = 0

    self._pi_network = policy
    self._critic_network = critic
    self._critic_opt = optax.sgd(learning_rate=critic_learning_rate)
    self._opponent_opt = optax.adam(opponent_model_learning_rate)
    self._policy_opt = optax.chain(
        optax.clip_by_global_norm(clip_grad_norm)
        if clip_grad_norm
        else optax.identity(),
        optax.sgd(learning_rate=pi_learning_rate),
    )
    self._train_state = self._init_train_state(info_state_size=info_state_size)
    self._current_policy = self.get_policy(return_probs=True)

    if correction_type == 'dice':
      policy_update_fn = get_dice_update_fn(
          agent_id=player_id,
          rng=self._rng,
          policy_network=policy,
          critic_network=critic,
          optimizer=self._policy_opt.update,
          opp_pi_lr=opp_policy_learning_rate,
          gamma=discount,
          n_lookaheads=n_lookaheads,
          env=env,
      )
    # pylint: disable=consider-using-in
    elif correction_type == 'lola' or correction_type == 'none':
      # if correction_type is none, use policy gradient without corrections
      lola_weight = 1.0 if correction_type == 'lola' else 0.0
      update_fn = get_lola_update_fn(
          agent_id=player_id,
          policy_network=policy,
          pi_lr=pi_learning_rate,
          optimizer=self._policy_opt.update,
          lola_weight=lola_weight,
      )
      policy_update_fn = jax.jit(update_fn) if use_jit else update_fn
    else:
      raise ValueError(f'Unknown correction type: {correction_type}')

    critic_update_fn = get_critic_update_fn(
        agent_id=player_id,
        critic_network=critic,
        optimizer=self._critic_opt.update,
        num_minibatches=num_critic_mini_batches,
        gamma=critic_discount,
    )

    self._policy_update_fns = {player_id: policy_update_fn}
    self._critic_update_fns = {
        player_id: jax.jit(critic_update_fn) if use_jit else critic_update_fn
    }

    for opponent in opponent_ids:
      opp_update_fn = get_opponent_update_fn(
          agent_id=opponent,
          policy_network=policy,
          optimizer=self._opponent_opt.update,
          num_minibatches=num_opponent_updates,
      )
      opp_critic_update_fn = get_critic_update_fn(
          agent_id=opponent,
          critic_network=critic,
          optimizer=self._critic_opt.update,
          num_minibatches=num_critic_mini_batches,
          gamma=critic_discount,
      )
      self._policy_update_fns[opponent] = (
          jax.jit(opp_update_fn) if use_jit else opp_update_fn
      )
      self._critic_update_fns[opponent] = (
          jax.jit(opp_critic_update_fn) if use_jit else opp_critic_update_fn
      )

  @property
  def train_state(self):
    return deepcopy(self._train_state)

  @property
  def policy_network(self):
    return self._pi_network

  @property
  def critic_network(self):
    return self._critic_network

  def metrics(self, return_last_only: bool = True):
    if not self._metrics:
      return {}
    metrics = self._metrics[-1] if return_last_only else self._metrics
    return metrics

  def update_params(self, state: TrainState, player_id: int) -> None:
    """Updates the parameters of the other agents.

    Args:
        state: the train state of the other agent.
        player_id: id of the other agent

    Returns:
    """
    self._train_state.policy_params[player_id] = deepcopy(
        state.policy_params[player_id]
    )
    self._train_state.critic_params[player_id] = deepcopy(
        state.critic_params[player_id]
    )

  def get_value_fn(self) -> typing.Callable:
    def value_fn(obs: jnp.ndarray):
      obs = jnp.array(obs)
      return self._critic_network.apply(
          self.train_state.critic_params[self.player_id], obs
      ).squeeze(-1)

    return jax.jit(value_fn)

  def get_policy(self, return_probs=True) -> typing.Callable:
    """Get the policy.
    
    Returns a function that takes a random key, an observation and
    optionally an action mask. The function produces actions which are
    sampled from the current policy. Additionally, if eturn_probs is true,
    it also returns the action probabilities.
    
    Args:
        return_probs: if true, the policy returns a tuple (action,
          action_probs).

    Returns:
        A function that maps observations to actions
    """

    def _policy(key: jax.random.PRNGKey, obs: jnp.ndarray, action_mask=None):
      """The actual policy function.
      
      Takes a random key, the current observation and optionally an action
      mask.
      
      Args:
          key: a random key for sampling
          obs: numpy array of observations
          action_mask: optional numpy array to mask out illegal actions

      Returns:
        Either the sampled actions or, if return_probs is true, a tuple
        (actions, action_probs).
      """
      params = self._train_state.policy_params[self.player_id]
      pi = self._pi_network.apply(params, obs)
      if action_mask is not None:
        probs = pi.probs * action_mask
        probs = probs / probs.sum()
        pi = distrax.Categorical(probs=probs)
      actions = pi.sample(seed=key)
      if return_probs:
        return actions, pi.prob(actions)
      else:
        return actions

    return jax.jit(_policy)

  def step(self, time_step: TimeStep, is_evaluation=False):
    """Produces an action and possibly triggers a parameter update.

    LOLA agents depend on having access to previous actions made by the
    opponent. Assumes that the field 'observations' of time_step contains a
    field 'actions' and its first axis is indexed by the player id. Similar, the
    fields 'rewards' and 'legal_actions' are assumed to be of shape
    (num_players,).

    Args:
        time_step: a TimeStep instance which has a field 'actions' in the
          observations dict.
        is_evaluation: if true, the agent will not update.

    Returns:
        A tuple containing the action that was taken and its probability
        under the current policy.
    """
    do_step = (
        time_step.is_simultaneous_move()
        or self.player_id == time_step.current_player()
    )
    action, probs = None, []
    batch_policy = vmap(self._current_policy, in_axes=(0, 0, None))
    if not time_step.last() and do_step:
      info_state = time_step.observations['info_state'][self.player_id]
      legal_actions = time_step.observations['legal_actions'][self.player_id]
      action_mask = np.zeros(self._num_actions)
      action_mask[legal_actions] = 1

      # If we are not in a batched environment, we need to add a batch dimension
      if 'batch_size' not in time_step.observations:
        info_state = jnp.array(info_state)[None]
        batch_size = 1
      else:
        batch_size = time_step.observations['batch_size']
      sample_keys = jax.random.split(next(self._rng), batch_size)
      action, probs = batch_policy(sample_keys, info_state, action_mask)

    if not is_evaluation:
      self._store_time_step(time_step=time_step, action=action)
      if time_step.last() and self._should_update():
        self._train_step()

    return rl_agent.StepOutput(action=action, probs=probs)

  def _init_train_state(self, info_state_size: chex.Shape):
    init_inputs = jnp.ones(info_state_size)
    agent_ids = self._opponent_ids + [self.player_id]
    policy_params, policy_opt_states = {}, {}
    critic_params, critic_opt_states = {}, {}
    for agent_id in agent_ids:
      policy_params[agent_id] = self._pi_network.init(
          next(self._rng), init_inputs
      )
      if agent_id == self.player_id:
        policy_opt_state = self._policy_opt.init(policy_params[agent_id])
      else:
        policy_opt_state = self._opponent_opt.init(policy_params[agent_id])
      policy_opt_states[agent_id] = policy_opt_state
      critic_params[agent_id] = self._critic_network.init(
          next(self._rng), init_inputs
      )
      critic_opt_states[agent_id] = self._critic_opt.init(
          critic_params[agent_id]
      )

    return TrainState(
        policy_params=policy_params,
        critic_params=critic_params,
        policy_opt_states=policy_opt_states,
        critic_opt_states=critic_opt_states,
    )

  def _store_time_step(self, time_step: TimeStep, action: np.ndarray):
    """Store the time step.
    
    Converts the timestep and the action into a transition and steps the
    counters.
    
    Args:
        time_step: the current time step.
        action: the action that was taken before observing time_step
    Returns: None
    """
    self._step_counter += (
        time_step.observations['batch_size']
        if 'batch_size' in time_step.observations
        else 1
    )
    if self._prev_time_step:
      transition = self._make_transition(time_step)
      self._data.append(transition)
    if time_step.last():
      self._prev_time_step = None
      self._prev_action = None
      self._episode_counter += 1
    else:
      obs = time_step.observations['info_state']
      time_step.observations['values'] = jnp.stack(
          [
              self._critic_network.apply(
                  self.train_state.critic_params[id], jnp.array(obs[id])
              ).squeeze(-1)
              for id in sorted(self.train_state.critic_params.keys())
          ]
      )
      self._prev_time_step = time_step
      self._prev_action = action

  def _train_step(self):
    """Updates the critic and the policy parameters.

    After the update, the data buffer is cleared. Returns: None
    """
    batch = self._construct_episode_batches(self._data)
    update_metrics = self._update_agent(batch)
    self._metrics.append(update_metrics)
    self._data.clear()

  def _should_update(self) -> bool:
    """Indicates whether to update or not.

    Returns:
        True, if the number of episodes in the buffer is equal to the batch
        size. False otherwise.
    """
    return (
        self._step_counter >= self._batch_size * (self._num_learn_steps + 1)
        and self._episode_counter > 0
    )

  def _update_agent(self, batch: TransitionBatch) -> typing.Dict:
    """Updates the critic and policy parameters of the agent.

    Args:
        batch: A batch of training episodes.
        
    Dimensions (N=player, B=batch_size, T=timesteps, S=state_dim):
      action: (N, B, T),
      discount: (B, T),
      info_state: (N, B, T, *S),
      legal_actions_mask: (N, B, T),
      reward: (N, B, T),
      terminal: (B, T),
      values: (N, B, T)

    Returns:
        A dictionary that contains relevant training metrics.
    """
    metrics = {}
    self._num_learn_steps += 1

    # if we do opponent modelling, we update the opponents first
    if self._fit_opponent_model:
      opponent_update_metrics = self._update_opponents(batch)
      metrics.update(
          (f'opp_models/{k}', v) for k, v in opponent_update_metrics.items()
      )

    # then we update the critic
    critic_update_metrics = self._update_critic(batch)
    metrics.update((f'critic/{k}', v) for k, v in critic_update_metrics.items())

    # and finally we update the policy
    if self._num_learn_steps % self._policy_update_interval == 0:
      policy_update_metrics = self._update_policy(batch)
      metrics.update(
          (f'policy/{k}', v) for k, v in policy_update_metrics.items()
      )
    return metrics

  def _construct_episode_batches(
      self, transitions: typing.List[TransitionBatch]
  ) -> TransitionBatch:
    """Constructs a list of transitions into a single transition batch instance.

    The fields 'info_state', 'rewards', 'legal_action_mask' and 'actions' of the
    produced transition batch have shape (num_agents, batch_size,
    sequence_length, *shape). The fields 'discount' and 'terminal' have shape
    (batch_size, sequence_length).

    Args:
        transitions: a list of single step transitions

    Returns:
        A transition batch instance with items of according shape.
    """
    episode, batches = [], []
    max_episode_length = 0
    for transition in transitions:
      episode.append(transition)
      if transition.terminal.any():
        max_episode_length = max(max_episode_length, len(episode))
        # pylint: disable=no-value-for-parameter
        batch = jax.tree_map(lambda *xs: jnp.stack(xs), *episode)
        batch = batch.replace(
            info_state=batch.info_state.transpose(1, 2, 0, 3),
            action=batch.action.transpose(1, 2, 0),
            legal_actions_mask=batch.legal_actions_mask.T,
            reward=batch.reward.transpose(1, 2, 0),
            values=batch.values.transpose(1, 2, 0),
            discount=batch.discount.transpose(1, 2, 0),
            terminal=batch.terminal.transpose(1, 2, 0),
        )
        batches.append(batch)
        episode.clear()
    return batches[0]

  def _update_policy(self, batch: TransitionBatch):
    self._train_state, metrics = self._policy_update_fns[self.player_id](
        self._train_state, batch
    )
    self._current_policy = self.get_policy(return_probs=True)
    return metrics

  def _update_critic(self, batch: TransitionBatch):
    self._train_state, metrics = self._critic_update_fns[self.player_id](
        self._train_state, batch
    )
    return metrics

  def _update_opponents(self, batch: TransitionBatch):
    update_metrics = {}
    for opponent in self._opponent_ids:
      self._train_state, metrics = self._critic_update_fns[opponent](
          self._train_state, batch
      )
      update_metrics.update(
          {f'agent_{opponent}/critic/{k}': v for k, v in metrics.items()}
      )
      self._train_state, metrics = self._policy_update_fns[opponent](
          self._train_state, batch
      )
      update_metrics.update(
          {f'agent_{opponent}/policy/{k}': v for k, v in metrics.items()}
      )
    return update_metrics

  def _make_transition(self, time_step: TimeStep):
    assert self._prev_time_step is not None
    legal_actions = self._prev_time_step.observations['legal_actions'][
        self.player_id
    ]
    legal_actions_mask = np.zeros((self._batch_size, self._num_actions))
    legal_actions_mask[..., legal_actions] = 1
    actions = np.array(time_step.observations['actions'])
    rewards = np.array(time_step.rewards)
    discounts = self._discount * (1 - time_step.last()) * np.ones_like(rewards)
    terminal = time_step.last() * np.ones_like(rewards)
    obs = np.array(self._prev_time_step.observations['info_state'])
    transition = TransitionBatch(
        info_state=obs,
        action=actions,
        reward=rewards,
        discount=discounts,
        terminal=terminal,
        legal_actions_mask=legal_actions_mask,
        values=self._prev_time_step.observations['values'],
    )
    if len(rewards.shape) < 2:  # if not a batch, add a batch dimension
      transition = jax.tree_map(lambda x: x[None], transition)
    return transition



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boltzmann Policy Iteration."""

from open_spiel.python import policy as policy_lib
from open_spiel.python.mfg.algorithms import mirror_descent


class BoltzmannPolicyIteration(mirror_descent.MirrorDescent):
  """Boltzmann Policy Iteration algorithm.

  In this algorithm, at each iteration, we update the policy by first computing
  the Q-function that evaluates the current policy, and then take a softmax.
  This corresponds to using Online Mirror Descent algorithm without summing
  Q-functions but simply taking the latest Q-function.
  """

  def get_projected_policy(self) -> policy_lib.Policy:
    """Returns the projected policy."""
    return mirror_descent.ProjectedPolicy(
        self._game,
        list(range(self._game.num_players())),
        self._state_value,
        coeff=self._lr)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.alpha_zero.evaluator."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import mcts
from open_spiel.python.algorithms.alpha_zero import evaluator as evaluator_lib
from open_spiel.python.algorithms.alpha_zero import model as model_lib
import pyspiel


def build_model(game):
  return model_lib.Model.build_model(
      "mlp", game.observation_tensor_shape(), game.num_distinct_actions(),
      nn_width=64, nn_depth=2, weight_decay=1e-4, learning_rate=0.01, path=None)


class EvaluatorTest(absltest.TestCase):

  def test_evaluator_caching(self):
    game = pyspiel.load_game("tic_tac_toe")
    model = build_model(game)
    evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)

    state = game.new_initial_state()
    obs = state.observation_tensor()
    act_mask = state.legal_actions_mask()
    action = state.legal_actions()[0]
    policy = np.zeros(len(act_mask), dtype=float)
    policy[action] = 1
    train_inputs = [model_lib.TrainInput(obs, act_mask, policy, value=1)]

    value = evaluator.evaluate(state)
    self.assertEqual(value[0], -value[1])
    value = value[0]

    value2 = evaluator.evaluate(state)[0]
    self.assertEqual(value, value2)

    prior = evaluator.prior(state)
    prior2 = evaluator.prior(state)
    np.testing.assert_array_equal(prior, prior2)

    info = evaluator.cache_info()
    self.assertEqual(info.misses, 1)
    self.assertEqual(info.hits, 3)

    for _ in range(20):
      model.update(train_inputs)

    # Still equal due to not clearing the cache
    value3 = evaluator.evaluate(state)[0]
    self.assertEqual(value, value3)

    info = evaluator.cache_info()
    self.assertEqual(info.misses, 1)
    self.assertEqual(info.hits, 4)

    evaluator.clear_cache()

    info = evaluator.cache_info()
    self.assertEqual(info.misses, 0)
    self.assertEqual(info.hits, 0)

    # Now they differ from before
    value4 = evaluator.evaluate(state)[0]
    value5 = evaluator.evaluate(state)[0]
    self.assertNotEqual(value, value4)
    self.assertEqual(value4, value5)

    info = evaluator.cache_info()
    self.assertEqual(info.misses, 1)
    self.assertEqual(info.hits, 1)

    value6 = evaluator.evaluate(game.new_initial_state())[0]
    self.assertEqual(value4, value6)

    info = evaluator.cache_info()
    self.assertEqual(info.misses, 1)
    self.assertEqual(info.hits, 2)

  def test_works_with_mcts(self):
    game = pyspiel.load_game("tic_tac_toe")
    model = build_model(game)
    evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)
    bot = mcts.MCTSBot(
        game, 1., 20, evaluator, solve=False, dirichlet_noise=(0.25, 1.))
    root = bot.mcts_search(game.new_initial_state())
    self.assertEqual(root.explore_count, 20)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for third_party.open_spiel.python.observation."""

import collections
import random
import time

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import get_all_states
from open_spiel.python.observation import INFO_STATE_OBS_TYPE
from open_spiel.python.observation import make_observation
import pyspiel


class ObservationTest(absltest.TestCase):

  def test_leduc_observation(self):
    game = pyspiel.load_game("leduc_poker")
    observation = make_observation(game)
    state = game.new_initial_state()
    state.apply_action(1)  # Deal 1
    state.apply_action(2)  # Deal 2
    state.apply_action(2)  # Bet
    state.apply_action(1)  # Call
    state.apply_action(3)  # Deal 3
    observation.set_from(state, player=0)
    np.testing.assert_array_equal(
        observation.tensor, [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 3, 3])
    self.assertEqual(
        list(observation.dict),
        ["player", "private_card", "community_card", "pot_contribution"])
    np.testing.assert_array_equal(observation.dict["player"], [1, 0])
    np.testing.assert_array_equal(observation.dict["private_card"],
                                  [0, 1, 0, 0, 0, 0])
    np.testing.assert_array_equal(observation.dict["community_card"],
                                  [0, 0, 0, 1, 0, 0])
    np.testing.assert_array_equal(observation.dict["pot_contribution"], [3, 3])
    self.assertEqual(
        observation.string_from(state, 0),
        "[Observer: 0][Private: 1][Round 2][Player: 0][Pot: 6]"
        "[Money: 97 97][Public: 3][Ante: 3 3]")

  def test_leduc_info_state(self):
    game = pyspiel.load_game("leduc_poker")
    observation = make_observation(game, INFO_STATE_OBS_TYPE)
    state = game.new_initial_state()
    state.apply_action(1)  # Deal 1
    state.apply_action(2)  # Deal 2
    state.apply_action(2)  # Bet
    state.apply_action(1)  # Call
    state.apply_action(3)  # Deal 3
    observation.set_from(state, player=0)
    np.testing.assert_array_equal(observation.tensor, [
        1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0
    ])
    self.assertEqual(
        list(observation.dict),
        ["player", "private_card", "community_card", "betting"])
    np.testing.assert_array_equal(observation.dict["player"], [1, 0])
    np.testing.assert_array_equal(observation.dict["private_card"],
                                  [0, 1, 0, 0, 0, 0])
    np.testing.assert_array_equal(observation.dict["community_card"],
                                  [0, 0, 0, 1, 0, 0])
    np.testing.assert_array_equal(
        observation.dict["betting"],
        [
            [[0, 1], [1, 0], [0, 0], [0, 0]],  # First round
            [[0, 0], [0, 0], [0, 0], [0, 0]],  # Second round
        ])
    self.assertEqual(
        observation.string_from(state, 0),
        "[Observer: 0][Private: 1][Round 2][Player: 0][Pot: 6]"
        "[Money: 97 97][Public: 3][Round1: 2 1][Round2: ]")

  def test_leduc_info_state_as_single_tensor(self):
    game = pyspiel.load_game("leduc_poker")
    observation = make_observation(
        game, INFO_STATE_OBS_TYPE,
        pyspiel.game_parameters_from_string("single_tensor"))
    state = game.new_initial_state()
    state.apply_action(1)  # Deal 1
    state.apply_action(2)  # Deal 2
    state.apply_action(2)  # Bet
    state.apply_action(1)  # Call
    state.apply_action(3)  # Deal 3
    observation.set_from(state, player=0)
    self.assertEqual(list(observation.dict), ["info_state"])
    np.testing.assert_array_equal(observation.dict["info_state"], [
        1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0
    ])

  def test_leduc_all_player_privates(self):
    game = pyspiel.load_game("leduc_poker")
    observation = make_observation(
        game,
        pyspiel.IIGObservationType(
            perfect_recall=True,
            private_info=pyspiel.PrivateInfoType.ALL_PLAYERS))
    state = game.new_initial_state()
    state.apply_action(1)  # Deal 1
    state.apply_action(2)  # Deal 2
    state.apply_action(2)  # Bet
    state.apply_action(1)  # Call
    state.apply_action(3)  # Deal 3
    observation.set_from(state, player=0)
    np.testing.assert_array_equal(observation.dict["private_cards"], [
        [0, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0],
    ])

  def test_benchmark_state_generation(self):
    # Generate trajectories to test on
    game = pyspiel.load_game("chess")
    trajectories = []
    for _ in range(20):
      state = game.new_initial_state()
      while not state.is_terminal():
        state.apply_action(random.choice(state.legal_actions()))
      trajectories.append(state.history())

    # New API
    total = 0
    observation = make_observation(game)
    start = time.time()
    for trajectory in trajectories:
      state = game.new_initial_state()
      for action in trajectory:
        state.apply_action(action)
        observation.set_from(state, 0)
        total += np.mean(observation.tensor)
    end = time.time()
    print("New API time per iteration "
          f"{1000*(end-start)/len(trajectories)}msec")

    # Old API
    total = 0
    start = time.time()
    for trajectory in trajectories:
      state = game.new_initial_state()
      for action in trajectory:
        state.apply_action(action)
        obs = state.observation_tensor(0)
        tensor = np.asarray(obs)
        total += np.mean(tensor)
    end = time.time()
    print("Old API time per iteration "
          f"{1000*(end-start)/len(trajectories)}msec")

  def test_compression_binary(self):
    # All infostates for leduc are binary, so we can compress them effectively.
    game = pyspiel.load_game("leduc_poker")
    obs1 = make_observation(game, INFO_STATE_OBS_TYPE)
    obs2 = make_observation(game, INFO_STATE_OBS_TYPE)
    self.assertLen(obs1.tensor, 30)  # 30 floats = 120 bytes
    for state in get_all_states.get_all_states(game).values():
      for player in range(game.num_players()):
        obs1.set_from(state, player)
        compressed = obs1.compress()
        self.assertEqual(type(compressed), bytes)
        self.assertLen(compressed, 5)
        obs2.decompress(compressed)
        np.testing.assert_array_equal(obs1.tensor, obs2.tensor)

  def test_compression_none(self):
    # Most observations for leduc have non-binary data, so we can't
    # currently compress them.
    game = pyspiel.load_game("leduc_poker")
    obs1 = make_observation(game)
    obs2 = make_observation(game)
    self.assertLen(obs1.tensor, 16)  # 16 floats = 64 bytes
    freq = collections.Counter()
    for state in get_all_states.get_all_states(game).values():
      for player in range(game.num_players()):
        obs1.set_from(state, player)
        compressed = obs1.compress()
        self.assertEqual(type(compressed), bytes)
        freq[len(compressed)] += 1
        obs2.decompress(compressed)
        np.testing.assert_array_equal(obs1.tensor, obs2.tensor)
    expected_freq = {
        3: 840,  # Compressible states take 3 bytes
        65: 17760,  # Uncompressible states take 65 bytes
    }
    self.assertEqual(freq, expected_freq)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for chess."""


from absl.testing import absltest
import numpy as np

import pyspiel
chess = pyspiel.chess


class GamesChessTest(absltest.TestCase):

  def test_bindings_sim(self):
    game = pyspiel.load_game("chess")
    state = game.new_initial_state()
    while not state.is_terminal():
      print(state)
      player = state.current_player()
      legal_actions = state.legal_actions()
      board = state.board()
      for action in legal_actions:
        action_str = state.action_to_string(player, action)
        move = chess.action_to_move(action, board)
        move_from = move.from_square
        move_to = move.to_square
        decoded_from_to = (f"({move_from.x} {move_from.y}) -> " +
                           f"({move_to.x} {move_to.y})")
        print(f"Legal action: {action_str} decoded from to {decoded_from_to}")
        print(f"Move representations: {move.to_string()} | " +
              f"{move.to_lan()} | {move.to_san(board)}")
      action = np.random.choice(legal_actions)
      state.apply_action(action)
    print(board.to_unicode_string())
    print(board.debug_string())
    print("Moves history:")
    print(" ".join([move.to_lan() for move in state.moves_history()]))
    self.assertTrue(state.is_terminal())


if __name__ == "__main__":
  np.random.seed(87375711)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset for structured payoff matrices."""

from absl import flags
import numpy as np

FLAGS = flags.FLAGS


class Dataset:
  """Dataset class."""

  def __init__(self, base_matrix, num_training_batches, minval, maxval):
    self._base_matrix = base_matrix
    self._num_training_batches = num_training_batches
    self._minval, self._maxval = minval, maxval
    # to overfit
    self._new_matrix = np.copy(self._base_matrix)

  def get_training_batch(self):
    """Get training data."""
    while True:
      if not FLAGS.single_problem:
        random_vec = np.random.randint(
            low=self._minval, high=self._maxval, size=FLAGS.batch_size)
        self._new_matrix = np.copy(self._base_matrix)
        for i in range(FLAGS.batch_size):
          self._new_matrix[self._new_matrix > 0] += random_vec[i]
          self._new_matrix[self._new_matrix < 0] -= random_vec[i]
      yield self._new_matrix

  def get_eval_batch(self):
    """Get eval dataset."""

    if not FLAGS.single_problem:
      random_vec = np.random.randint(
          low=self._minval, high=self._maxval, size=FLAGS.batch_size)
      self._new_matrix = np.copy(self._base_matrix)
      for i in range(FLAGS.batch_size):
        self._new_matrix[self._new_matrix > 0] += random_vec[i]
        self._new_matrix[self._new_matrix < 0] -= random_vec[i]
    return self._new_matrix



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plays a uniform random bot against the default scenarios for that game."""

import random
from absl import app
from absl import flags

from open_spiel.python.bots import scenarios
from open_spiel.python.bots import uniform_random
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_string("game_name", "catch", "Game to play scenarios for.")


def main(argv):
  del argv
  game = pyspiel.load_game(FLAGS.game_name)

  # TODO(author1): Add support for bots from neural networks.
  bots = [
      uniform_random.UniformRandomBot(i, random)
      for i in range(game.num_players())
  ]
  scenarios.play_bot_in_scenarios(game, bots)


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RNN meta-regret matching with self-play agents."""

from typing import List

from absl import flags
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python.examples.meta_cfr.matrix_games.rnn_model import RNNModel

FLAGS = flags.FLAGS


def _make_network(lstm_hidden_sizes: List[int],
                  mlp_hidden_sizes: List[int],
                  output_dim: int) -> hk.RNNCore:
  """set up the network."""

  layers = []
  for k, hidden_size in enumerate(lstm_hidden_sizes):
    layers += [hk.LSTM(hidden_size, name=f'lstm_layer_{k}'), jax.nn.relu]
  layers += [hk.nets.MLP(mlp_hidden_sizes + [output_dim], name='mlp')]
  return RNNModel(layers)


def _make_forwards(lstm_hidden_sizes: List[int], mlp_hidden_sizes: List[int],
                   output_dim: int, batch_size: int) -> hk.Transformed:

  """Forward pass."""

  def forward_fn(inputs):
    rnn = _make_network(lstm_hidden_sizes, mlp_hidden_sizes, output_dim)
    initial_state = rnn.initial_state(batch_size=batch_size)
    outputs, _ = hk.dynamic_unroll(rnn, inputs, initial_state, time_major=False)
    return outputs

  network = hk.transform(forward_fn)
  return network


def meta_loss(opt_params, net_apply, payoff, steps, rng):
  """Meta loss function."""

  regret_sum_x = np.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])
  regret_sum_y = np.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])
  total_loss = 0

  @jax.jit
  def body_fun(s, total_loss):
    nonlocal regret_sum_x
    nonlocal regret_sum_y
    x = net_apply(opt_params, rng, regret_sum_x / (s + 1))
    y = net_apply(opt_params, rng, regret_sum_y / (s + 1))

    strategy_x = jax.nn.softmax(x)
    strategy_y = jnp.transpose(jax.nn.softmax(y), [0, 2, 1])

    values_x = jnp.matmul(payoff, strategy_y)
    values_y = -jnp.matmul(strategy_x, payoff)

    value_x = jnp.matmul(jnp.matmul(strategy_x, payoff), strategy_y)
    value_y = -value_x

    curren_regret_x = values_x - value_x
    curren_regret_y = values_y - value_y
    curren_regret_x = jnp.transpose(curren_regret_x, [0, 2, 1])

    regret_sum_x += curren_regret_x
    regret_sum_y += curren_regret_y

    current_loss = jnp.max(
        jax.numpy.concatenate([curren_regret_x, curren_regret_y], axis=2),
        axis=[1, 2])
    total_loss += current_loss
    return total_loss
  def fori_loop(lower, steps, body_fun, total_loss):
    val = total_loss
    for i in range(lower, steps):
      val = body_fun(i, total_loss)
    return val
  total_loss = fori_loop(0, steps, body_fun, total_loss)
  return jnp.mean(total_loss)


class OptimizerModel:
  """Optimizer model."""

  def __init__(self, learning_rate):
    self.learning_rate = learning_rate
    self.model = _make_forwards(
        lstm_hidden_sizes=[20],
        mlp_hidden_sizes=[],
        output_dim=3,
        batch_size=FLAGS.batch_size)
    self.net_apply = self.model.apply
    self.net_init = self.model.init
    self.opt_update, self.net_params, self.opt_state = None, None, None

  def lr_scheduler(self, init_value):
    schedule_fn = optax.polynomial_schedule(
        init_value=init_value, end_value=0.05, power=1., transition_steps=50)
    return schedule_fn

  def get_optimizer_model(self):
    schedule_fn = self.lr_scheduler(self.learning_rate)
    opt_init, self.opt_update = optax.chain(
        optax.scale_by_adam(), optax.scale_by_schedule(schedule_fn),
        optax.scale(-self.learning_rate))
    rng = jax.random.PRNGKey(10)
    dummy_input = np.random.normal(
        loc=0, scale=10., size=(FLAGS.batch_size, 1, FLAGS.num_actions))
    self.net_params = self.net_init(rng, dummy_input)
    self.opt_state = opt_init(self.net_params)


class MetaSelfplayAgent:
  """Meta player agent."""

  def __init__(self, repeats, training_epochs, data_loader):
    self.repeats = repeats
    self.training_epochs = training_epochs
    self.net_apply = None
    self.net_params = None
    self.regret_sum = None
    self.step = 0
    self.data_loader = data_loader
    self._rng = hk.PRNGSequence(10)

  def train(self):
    self.training_optimizer()
    self.regret_sum = jnp.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])

  def initial_policy(self):
    x = self.net_apply(self.net_params, next(self._rng), self.regret_sum)
    self.last_policy = jax.nn.softmax(x)
    self.step += 1
    return self.last_policy

  def next_policy(self, last_values):
    value = jnp.matmul(self.last_policy, last_values)
    curren_regret = jnp.transpose(last_values, [0, 2, 1]) - value
    self.regret_sum += curren_regret

    x = self.net_apply(self.net_params, next(self._rng),
                       self.regret_sum / (self.step + 1))
    self.last_policy = jax.nn.softmax(x)
    self.step += 1
    return self.last_policy

  def training_optimizer(self):
    """Train optimizer."""

    optimizer = OptimizerModel(0.01)
    optimizer.get_optimizer_model()
    for _ in range(FLAGS.num_batches):
      batch_payoff = next(self.data_loader)
      for _ in range(self.repeats):
        grads = jax.grad(
            meta_loss, has_aux=False)(optimizer.net_params, optimizer.net_apply,
                                      batch_payoff, self.training_epochs,
                                      next(self._rng))

        updates, optimizer.opt_state = optimizer.opt_update(
            grads, optimizer.opt_state)
        optimizer.net_params = optax.apply_updates(optimizer.net_params,
                                                   updates)
    self.net_apply = optimizer.net_apply
    self.net_params = optimizer.net_params
